#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.index module

"""

from hypatia.keyword import KeywordIndex

from pyams_security.interfaces import IProtectedObject


__docformat__ = 'restructuredtext'


class PrincipalsRoleIndex(KeywordIndex):
    """Principals role index"""

    def __init__(self, role_id, family=None):
        KeywordIndex.__init__(self, role_id, family)
        self.role_id = role_id

    def discriminate(self, obj, default):
        protected_object = IProtectedObject(obj, None)
        if protected_object is None:
            return default
        return protected_object.get_principals(self.role_id)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces.names module

Package constant strings.
"""

__docformat__ = 'restructuredtext'

from pyams_security import _


SYSTEM_PREFIX = 'system'
ADMIN_USER_NAME = '__system__'
ADMIN_USER_LOGIN = 'admin'
ADMIN_USER_ID = '{0}:{1}'.format(SYSTEM_PREFIX, ADMIN_USER_LOGIN)

INTERNAL_USER_NAME = '__internal__'
INTERNAL_USER_LOGIN = 'internal'
INTERNAL_USER_ID = '{0}:{1}'.format(SYSTEM_PREFIX, INTERNAL_USER_LOGIN)

SYSTEM_ADMIN_ROLE = 'system.Manager'
SYSTEM_VIEWER_ROLE = 'system.Viewer'

USER_LOGIN_TITLE = _("User login")

UNKNOWN_PRINCIPAL_ID = '__UNKNOWN__'
UNCHANGED_PASSWORD = '*****'

PRINCIPAL_ID_FORMATTER = '{prefix}:{login}'
GROUP_ID_FORMATTER = '{prefix}:{group_id}'


PERMISSIONS_VOCABULARY_NAME = 'pyams_security.permissions'
ROLES_VOCABULARY_NAME = 'pyams_security.roles'
PASSWORD_MANAGERS_VOCABULARY_NAME = 'pyams_security.password.managers'


USERS_FOLDERS_VOCABULARY_NAME = 'pyams_security.plugin.users-folders'
LOCAL_GROUPS_VOCABULARY_NAME = 'pyams_security.plugin.local-groups'



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.schema module

This module contains security-related schema fields.
"""

from zope.interface import Interface, implementer
from zope.schema import Choice, Set, TextLine
from zope.schema.interfaces import IChoice, ISet, ITextLine

from pyams_security.interfaces.base import IPermission, IRole, IPrincipalInfo
from pyams_security.interfaces.names import PERMISSIONS_VOCABULARY_NAME, ROLES_VOCABULARY_NAME


__docformat__ = 'restructuredtext'


class IPermissionField(IChoice):
    """Permission field interface"""


@implementer(IPermissionField)
class PermissionField(Choice):
    """Permission field"""

    def __init__(self, **kwargs):
        if 'vocabulary' in kwargs:
            del kwargs['vocabulary']
        super().__init__(vocabulary=PERMISSIONS_VOCABULARY_NAME, **kwargs)

    def validate(self, value):
        if IPermission.providedBy(value):
            value = value.id
        super().validate(value)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if IPermission.providedBy(value):
            value = value.id
        super().set(object, value)


class IPermissionsSetField(ISet):
    """Permissions set field interface"""


def get_permission_id(value):
    """Get permission ID"""
    return value.id if IPermission.providedBy(value) else value


@implementer(IPermissionsSetField)
class PermissionsSetField(Set):
    """Permissions set field"""

    value_type = PermissionField()

    def __init__(self, **kwargs):
        if 'value_type' in kwargs:
            del kwargs['value_type']
        super().__init__(value_type=PermissionField(), **kwargs)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if value:
            value = set(map(get_permission_id, value))
        super().set(object, value)


class IRoleField(IChoice):
    """Role field interface"""


@implementer(IRoleField)
class RoleField(Choice):
    """Role field"""

    def __init__(self, **kwargs):
        if 'vocabulary' in kwargs:
            del kwargs['vocabulary']
        super().__init__(vocabulary=ROLES_VOCABULARY_NAME, **kwargs)

    def validate(self, value):
        if IRole.providedBy(value):
            value = value.id
        super().validate(value)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if IRole.providedBy(value):
            value = value.id
        super().set(object, value)


class IRolesSetField(ISet):
    """Roles set field interface"""


def get_role_id(value):
    """Get role ID"""
    return value.id if IRole.providedBy(value) else value


@implementer(IRolesSetField)
class RolesSetField(Set):
    """Roles set field"""

    value_type = RoleField()

    def __init__(self, **kwargs):
        if 'value_type' in kwargs:
            del kwargs['value_type']
        super().__init__(value_type=RoleField(), **kwargs)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if value:
            value = set(map(get_role_id, value))
        super().set(object, value)


class IPrincipalBaseField(Interface):
    """Base role field interface"""

    role_id = TextLine(title="Matching role ID",
                       required=False)


class IPrincipalField(ITextLine, IPrincipalBaseField):
    """Principal field interface"""


@implementer(IPrincipalField)
class PrincipalField(TextLine):
    """Principal field"""

    role_id = None

    def __init__(self, **kwargs):
        if 'role_id' in kwargs:
            self.role_id = kwargs.pop('role_id')
        super().__init__(**kwargs)

    def validate(self, value):
        if IPrincipalInfo.providedBy(value):
            value = value.id
        super().validate(value)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if IPrincipalInfo.providedBy(value):
            value = value.id
        super().set(object, value)


class IPrincipalsSetField(ISet, IPrincipalBaseField):
    """Principals set interface"""


def get_principal_id(value):
    """Get principal ID"""
    return value.id if IPrincipalInfo.providedBy(value) else value


@implementer(IPrincipalsSetField)
class PrincipalsSetField(Set):
    """Principals set field"""

    role_id = None
    value_type = PrincipalField()

    def __init__(self, **kwargs):
        if 'role_id' in kwargs:
            self.role_id = kwargs.pop('role_id')
        super().__init__(**kwargs)

    def set(self, object, value):  # pylint: disable=redefined-builtin
        if value:
            value = set(map(get_principal_id, value))
        super().set(object, value)



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.rest module

This module defines components which are used to improve security of REST services
using CORS requests.
"""

import sys

from persistent import Persistent
from pyramid.httpexceptions import HTTPBadRequest, HTTPServerError
from zope.container.contained import Contained
from zope.schema.fieldproperty import FieldProperty

from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.rest import CORS_CONFIGURATION_KEY, ICORSSecurityInfo
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.interfaces.rest import ICORSRequestHandler
from pyams_utils.registry import get_utility
from pyams_utils.rest import CORSRequestHandler


__docformat__ = 'restructuredtext'


@factory_config(ICORSSecurityInfo)
class CORSSecurityInfo(Persistent, Contained):
    """CORS security persistent info"""

    restrict_origins = FieldProperty(ICORSSecurityInfo['restrict_origins'])
    allowed_origins = FieldProperty(ICORSSecurityInfo['allowed_origins'])

    def check_origin(self, request):
        """Check origin of CORS request"""
        if not self.restrict_origins:
            return
        origin = request.headers.get('Origin', request.host_url)
        if (origin == request.host_url) or (origin in (self.allowed_origins or ())):
            return
        raise HTTPBadRequest('Forbidden origin')

    @staticmethod
    def set_headers(request, allowed_methods=None):
        """Set CORS headers of provided request"""
        req_headers = request.headers
        resp_headers = request.response.headers
        resp_headers['Access-Control-Allow-Credentials'] = 'true'
        resp_headers['Access-Control-Allow-Origin'] = \
            req_headers.get('Origin', request.host_url)
        if 'Access-Control-Request-Headers' in req_headers:
            resp_headers['Access-Control-Allow-Headers'] = \
                req_headers.get('Access-Control-Request-Headers', 'Origin')
        if 'Access-Control-Request-Method' in req_headers:
            try:
                service = request.current_service
                resp_headers['Access-Control-Allow-Methods'] = \
                    ', '.join(service.cors_supported_methods)
            except AttributeError as exc:
                if allowed_methods:
                    resp_headers['Access-Control-Allow-Methods'] = ', '.join(allowed_methods)
                else:
                    test_mode = sys.argv[-1].endswith('/test')
                    if not test_mode:
                        raise HTTPServerError from exc


@adapter_config(required=ISecurityManager, provides=ICORSSecurityInfo)
def security_manager_cors_configuration_factory(context):
    """Security manager CORS configuration factory"""
    return get_annotation_adapter(context, CORS_CONFIGURATION_KEY, ICORSSecurityInfo)


def check_cors_origin(request, **kwargs):  # pylint: disable=unused-argument
    """Cornice service origin validator"""
    sm = get_utility(ISecurityManager)  # pylint: disable=invalid-name
    cors_info = ICORSSecurityInfo(sm)
    cors_info.check_origin(request)


def set_cors_headers(request, **kwargs):  # pylint: disable=unused-argument
    """Set REST CORS headers"""
    sm = get_utility(ISecurityManager)  # pylint: disable=invalid-name
    cors_info = ICORSSecurityInfo(sm)
    cors_info.set_headers(request, allowed_methods=kwargs.get('allowed_methods'))


@adapter_config(required=IPyAMSLayer,
                provides=ICORSRequestHandler)
class ProtectedCORSRequestHandler(CORSRequestHandler):
    """Protected CORS request handler"""

    def handle_request(self, allowed_methods=None):
        """Check request origin and add requested headers to current request"""
        check_cors_origin(self.request)
        set_cors_headers(self.request, allowed_methods=allowed_methods)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.principal module

This module provides principal related classes.
"""

from zope.annotation.interfaces import IAnnotations
from zope.interface import implementer
from zope.principalannotation.interfaces import IPrincipalAnnotationUtility
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces.base import IPrincipalInfo
from pyams_security.interfaces.names import UNKNOWN_PRINCIPAL_ID
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import query_utility


__docformat__ = 'restructuredtext'

from pyams_security import _  # pylint: disable=ungrouped-imports


@implementer(IPrincipalInfo)
class PrincipalInfo:
    """Generic principal info"""

    id = FieldProperty(IPrincipalInfo['id'])  # pylint: disable=invalid-name
    title = FieldProperty(IPrincipalInfo['title'])
    attributes = FieldProperty(IPrincipalInfo['attributes'])

    def __init__(self, **kwargs):
        self.id = kwargs.pop('id')  # pylint: disable=invalid-name
        self.title = kwargs.pop('title', '__unknown__')
        self.attributes = kwargs

    def __eq__(self, other):
        return isinstance(other, PrincipalInfo) and (self.id == other.id)

    def __hash__(self):
        return hash(self.id)


@implementer(IPrincipalInfo)
class UnknownPrincipal:
    """Unknown principal info"""

    id = UNKNOWN_PRINCIPAL_ID  # pylint: disable=invalid-name
    title = _("< unknown principal >")


UnknownPrincipal = UnknownPrincipal()  # pylint: disable=invalid-name


@implementer(IPrincipalInfo)
class MissingPrincipal:
    """Missing principal info

    This class can be used when a stored principal ID
    references a principal which can't be found anymore
    """

    id = FieldProperty(IPrincipalInfo['id'])  # pylint: disable=invalid-name

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')  # pylint: disable=invalid-name

    @property
    def title(self):
        """Get principal title"""
        return 'MissingPrincipal: {id}'.format(id=self.id)

    def __eq__(self, other):
        return isinstance(other, PrincipalInfo) and (self.id == other.id)


@adapter_config(required=IPrincipalInfo, provides=IAnnotations)
def get_principal_annotations(principal):
    """Principal annotations adapter"""
    annotations = query_utility(IPrincipalAnnotationUtility)
    if annotations is not None:
        return annotations.getAnnotations(principal)
    return None



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.role module

This module provides classes related to roles definition and registration.
"""
from pyramid.exceptions import ConfigurationError
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_security.interfaces import IRoleEvent
from pyams_security.interfaces.base import IRole
from pyams_security.interfaces.names import ROLES_VOCABULARY_NAME
from pyams_utils.request import check_request
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'


@implementer(IRole)
class Role:
    """Role utility class"""

    id = FieldProperty(IRole['id'])  # pylint: disable=invalid-name
    title = FieldProperty(IRole['title'])
    description = FieldProperty(IRole['description'])
    permissions = FieldProperty(IRole['permissions'])
    managers = FieldProperty(IRole['managers'])

    def __init__(self, values=None, **args):
        if not isinstance(values, dict):
            values = args
        self.id = values.get('id')  # pylint: disable=invalid-name
        self.title = values.get('title')
        self.description = values.get('description')
        self.permissions = values.get('permissions')
        self.managers = values.get('managers')


class RoleSelector:
    """Role based event selector predicate

    This selector can be used as a subscriber predicate to define
    a role that the event must match:

    .. code-block:: python

        from pyams_utils.interfaces.site import ISiteRoot

        @subscriber(IRoleGrantedEvent, context_selector=ISiteRoot, role_selector='myams.admin')
        def handle_granted_manager_role(event):
            '''Handle granted manager role on site root'''
    """

    def __init__(self, roles, config):  # pylint: disable=unused-argument
        if not isinstance(roles, (list, tuple, set)):
            roles = {roles}
        self.roles = roles

    def text(self):
        """Predicate text output"""
        return 'role_selector = %s' % str(self.roles)

    phash = text

    def __call__(self, event):
        assert IRoleEvent.providedBy(event)
        return event.role_id in self.roles


def register_role(config, role):
    """Register a new role

    Roles registry is not required.
    But only registered roles can be applied via default
    ZMI features.

    If a role is registered several times, previous registrations
    will just be updated to add new permissions.
    Title and description are not updated after first registration.
    """
    registry = config.registry
    if not IRole.providedBy(role):
        if isinstance(role, dict):
            role = Role(**role)
        else:
            role = Role(id=role, title=role)
    role_utility = registry.queryUtility(IRole, name=role.id)
    if role_utility is None:
        # registering a new role
        registry.registerUtility(role, IRole, name=role.id)
    else:
        # new registration of a given role to add permissions
        role_utility.permissions = (role_utility.permissions or set()) | \
                                   (role.permissions or set())
        role_utility.managers = (role_utility.managers or set()) | \
                                (role.managers or set())


def upgrade_role(config, role, required=True, **kwargs):
    """Upgrade an existing role with new permissions or managers

    If *required* argument is True, an exception is raised if provided
    role doesn't exist.
    """
    registry = config.registry
    if not IRole.providedBy(role):
        role = registry.queryUtility(IRole, name=role)
    if role is None:
        if required:
            raise ConfigurationError("Provided role isn't registered!")
        return
    if 'permissions' in kwargs:
        role.permissions = (role.permissions or set()) | set(kwargs['permissions'])
    if 'managers' in kwargs:
        role.managers = (role.managers or set()) | set(kwargs['managers'])


@vocabulary_config(name=ROLES_VOCABULARY_NAME)
class RolesVocabulary(SimpleVocabulary):
    """Roles vocabulary"""

    interface = IRole

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        request = check_request()
        registry = request.registry
        translate = request.localizer.translate
        terms = [SimpleTerm(r.id, title=translate(r.title))
                 for n, r in registry.getUtilitiesFor(self.interface)]
        terms.sort(key=lambda x: x.title)
        super().__init__(terms)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.property module

This module defines a custom field property used to store roles.
"""

from zope.schema.interfaces import IField, ISet

from pyams_security.interfaces import IProtectedObject, IRoleProtectedObject
from pyams_security.interfaces.base import IRole, IPrincipalInfo
from pyams_security.schema import IPrincipalBaseField


__docformat__ = 'restructuredtext'


class RolePrincipalsFieldProperty:
    """Custom field property used to handle role principals"""

    def __init__(self, field, role_id=None, name=None, **args):  # pylint: disable=unused-argument
        if not IField.providedBy(field):
            raise ValueError("Provided field must implement IField interface")
        if role_id is None:
            if not IPrincipalBaseField.providedBy(field):
                raise ValueError("Provided field must implement IRoleField interface "
                                 "or you must provide a role ID")
            role_id = field.role_id
        elif IRole.providedBy(role_id):
            role_id = role_id.id
        if role_id is None:
            raise ValueError("Can't get role ID")
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        self.__role_id = role_id

    def __get__(self, instance, klass):
        if instance is None:
            return self
        protection = IProtectedObject(instance.__parent__, None)
        if protection is None:
            return set()
        return protection.get_principals(self.__role_id)

    def __set__(self, instance, value):
        field = self.__field.bind(instance)
        if ISet.providedBy(field):  # pylint: disable=no-value-for-parameter
            if value is None:
                value = set()
            elif isinstance(value, str):
                value = set(value.split(','))
            value = set(map(lambda x: x.id if IPrincipalInfo.providedBy(x) else x, value))
        else:
            value = value.id if IPrincipalInfo.providedBy(value) else value
        field.validate(value)
        if field.readonly:
            raise ValueError("Field {0} is readonly!".format(self.__name))
        protection = IProtectedObject(instance.__parent__, None)
        if not IRoleProtectedObject.providedBy(protection):
            raise ValueError("Can't use role properties on object not providing "
                             "IRoleProtectedObject interface!")
        # pylint: disable=assignment-from-no-return
        old_principals = protection.get_principals(self.__role_id)
        if not isinstance(value, set):
            value = {value}
        added = value - old_principals
        removed = old_principals - value
        for principal_id in added:
            protection.grant_role(self.__role_id, principal_id)
        for principal_id in removed:
            protection.revoke_role(self.__role_id, principal_id)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

import logging

from zope.annotation import IAnnotations

from pyams_security.interfaces import IDefaultProtectionPolicy
from pyams_security.security import POLICY_ANNOTATIONS_KEY
from pyams_utils.container import find_objects_providing
from pyams_utils.registry import get_local_registry, set_local_registry


LOGGER = logging.getLogger('PyAMS (security)')

ROLES_ANNOTATIONS_KEY = 'pyams_security.roles'


def evolve(site):
    """Evolve 1: update roles annotations"""
    registry = get_local_registry()
    try:
        set_local_registry(site.getSiteManager())
        for content in find_objects_providing(site, IDefaultProtectionPolicy):
            annotations = IAnnotations(content)
            if ROLES_ANNOTATIONS_KEY in annotations:
                LOGGER.info("Updating roles annotations for {!r}".format(content))
                annotations[POLICY_ANNOTATIONS_KEY] = annotations[ROLES_ANNOTATIONS_KEY]
                del annotations[ROLES_ANNOTATIONS_KEY]
    finally:
        set_local_registry(registry)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.interfaces.profile module

"""

from zope.annotation.interfaces import IAttributeAnnotatable

from pyams_file.schema import ThumbnailImageField


__docformat__ = 'restructuredtext'

from pyams_site import _


PUBLIC_PROFILE_KEY = 'pyams_security.public_profile'


class IPublicProfile(IAttributeAnnotatable):
    """User public profile preferences"""

    avatar = ThumbnailImageField(title=_("Profile's avatar"),
                                 description=_("This picture will be associated to your user "
                                               "profile"),
                                 required=False)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.plugin.group module

This module defines local groups of principals.
"""

import logging
from BTrees import OOBTree  # pylint: disable=no-name-in-module
from persistent import Persistent
from pyramid.events import subscriber
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.names import GROUP_ID_FORMATTER, LOCAL_GROUPS_VOCABULARY_NAME
from pyams_security.interfaces.plugin import IGroupsFolderPlugin, ILocalGroup, \
    IPrincipalsAddedToGroupEvent, IPrincipalsRemovedFromGroupEvent, \
    PrincipalsAddedToGroupEvent, PrincipalsRemovedFromGroupEvent
from pyams_security.principal import PrincipalInfo
from pyams_utils.factory import factory_config
from pyams_utils.registry import query_utility
from pyams_utils.request import check_request
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'

LOGGER = logging.getLogger('PyAMS(security)')


@factory_config(ILocalGroup)
class Group(Persistent, Contained):
    """Local group persistent class"""

    group_id = FieldProperty(ILocalGroup['group_id'])
    title = FieldProperty(ILocalGroup['title'])
    description = FieldProperty(ILocalGroup['description'])

    _principals = FieldProperty(ILocalGroup['principals'])

    @property
    def principals(self):
        """Get principals list"""
        return self._principals or set()

    @principals.setter
    def principals(self, value):
        if not value:
            value = set()
        added = value - self._principals
        removed = self._principals - value
        if added or removed:
            self._principals = value
            registry = check_request().registry
            if added:
                LOGGER.debug("Added principals {0} to group {1} ({2})".format(
                    str(added), self.group_id, self.title))
                registry.notify(PrincipalsAddedToGroupEvent(self, added))
            if removed:
                LOGGER.debug("Removed principals {0} from group {1} ({2})".format(
                    str(removed), self.group_id, self.title))
                registry.notify(PrincipalsRemovedFromGroupEvent(self, removed))


@vocabulary_config(name=LOCAL_GROUPS_VOCABULARY_NAME)
class LocalGroupsVocabulary(SimpleVocabulary):
    """'PyAMS local groups vocabulary"""

    def __init__(self, context=None):  # pylint: disable=unused-argument
        terms = []
        manager = query_utility(ISecurityManager)
        if manager is not None:
            for plugin in manager.values():
                if IGroupsFolderPlugin.providedBy(plugin):
                    for group in plugin.values():
                        terms.append(SimpleTerm(GROUP_ID_FORMATTER.format(
                            prefix=plugin.prefix, group_id=group.group_id),
                                                title=group.title))
        super().__init__(terms)


@factory_config(IGroupsFolderPlugin)
class GroupsFolder(Folder):
    """Principals groups folder"""

    prefix = FieldProperty(IGroupsFolderPlugin['prefix'])
    title = FieldProperty(IGroupsFolderPlugin['title'])
    enabled = FieldProperty(IGroupsFolderPlugin['enabled'])

    def __init__(self):
        super().__init__()
        self.groups_by_principal = OOBTree.OOBTree()  # pylint: disable=no-member

    def check_group_id(self, group_id):
        """Check for existence of given group ID"""
        if not group_id:
            return False
        return group_id not in self

    def get_principal(self, principal_id, info=True):
        """Principal lookup for given principal ID"""
        if not self.enabled:
            return None
        if not principal_id.startswith(self.prefix + ':'):
            return None
        prefix, group_id = principal_id.split(':', 1)  # pylint: disable=unused-variable
        group = self.get(group_id)
        if group is not None:
            if info:
                return PrincipalInfo(id=GROUP_ID_FORMATTER.format(prefix=self.prefix,
                                                                  group_id=group.group_id),
                                     title=group.title)
            return group
        return None

    def get_all_principals(self, principal_id, seen=None):
        """Get all principals matching given principal ID"""
        if not self.enabled:
            return set()
        principals = self.groups_by_principal.get(principal_id) or set()
        principals = principals.copy()
        if principals:
            if seen is None:
                seen = set()
            for principal in (p for p in principals.copy() if p not in seen):
                seen.add(principal)
                if principal.startswith(self.prefix + ':'):
                    principals.update(self.get_all_principals(principal, seen))
        return principals

    def find_principals(self, query, exact_match=False):
        """Find principals matching given query"""
        if not self.enabled:
            return
        if not query:
            return
        query = query.lower()
        for group in self.values():
            title = group.title
            if not title:
                continue
            title = title.lower()
            if (query == title) or (not exact_match and query in title):
                yield PrincipalInfo(id=GROUP_ID_FORMATTER.format(prefix=self.prefix,
                                                                 group_id=group.group_id),
                                    title=group.title)


@subscriber(IObjectAddedEvent, context_selector=ILocalGroup)
def handle_added_group(event):
    """Handle added group"""
    group = event.object
    folder = event.newParent
    principals_map = folder.groups_by_principal
    for principal_id in group.principals:
        groups_set = principals_map.get(principal_id)
        if groups_set is None:
            groups_set = set()
        group_id = GROUP_ID_FORMATTER.format(prefix=folder.prefix,
                                             group_id=group.group_id)
        groups_set.add(group_id)
        principals_map[principal_id] = groups_set


@subscriber(IPrincipalsAddedToGroupEvent)
def handle_added_principals(event):
    """Handle principals added to group"""
    group = event.group
    if group.__parent__ is None:  # can occur when a group is created
        return
    principals_map = group.__parent__.groups_by_principal
    for principal_id in event.principals:
        groups_set = principals_map.get(principal_id)
        if groups_set is None:
            groups_set = set()
        group_id = GROUP_ID_FORMATTER.format(prefix=group.__parent__.prefix,
                                             group_id=group.group_id)
        groups_set.add(group_id)
        principals_map[principal_id] = groups_set


@subscriber(IPrincipalsRemovedFromGroupEvent)
def handle_removed_principals(event):
    """Handle principals removed from group"""
    group = event.group
    principals_map = group.__parent__.groups_by_principal
    for principal_id in event.principals:
        groups_set = principals_map.get(principal_id)
        if groups_set:
            group_id = GROUP_ID_FORMATTER.format(prefix=group.__parent__.prefix,
                                                 group_id=group.group_id)
            if group_id in groups_set:
                groups_set.remove(group_id)
            if groups_set:
                principals_map[principal_id] = groups_set
            else:
                del principals_map[principal_id]



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces.base module

This module defines base security permissions and interfaces.
"""

from zope.interface import Interface
from zope.schema import Dict, Set, Text, TextLine


__docformat__ = 'restructuredtext'


FORBIDDEN_PERMISSION = 'system.forbidden'
'''Custom permission which is never granted to any user, even system manager'''

PUBLIC_PERMISSION = 'public'
'''Public permission which is granted to every principal'''

VIEW_PERMISSION = 'view'
'''View permission is a custom permission used to view contents'''

MANAGE_PERMISSION = 'manage'
'''Permission used to manage basic information; this permission is generally not used by custom
contents'''

VIEW_SYSTEM_PERMISSION = 'pyams.ViewSystem'
'''Permission used to access management screens'''

MANAGE_SYSTEM_PERMISSION = 'pyams.ManageSystem'
'''Permission used to manage system settings'''

MANAGE_SECURITY_PERMISSION = 'pyams.ManageSecurity'
'''Permission used to manage security settings'''

MANAGE_ROLES_PERMISSION = 'pyams.ManageRoles'
'''Permission used to manage roles'''


class IPermission(Interface):
    """Permission interface"""

    # pylint: disable=invalid-name
    id = TextLine(title="Unique ID",
                  required=True)

    title = TextLine(title="Title",
                     required=True)

    description = Text(title="Description",
                       required=False)


ROLE_ID = 'role:{}'
'''Role ID formatter string'''


class IRole(Interface):
    """Role interface

    A role is a set of permissions; by assigning the role to a principal,
    these permissions are also granted to the principal.
    """

    # pylint: disable=invalid-name
    id = TextLine(title="Unique ID",
                  required=True)

    title = TextLine(title="Title",
                     required=True)

    description = Text(title="Description",
                       required=False)

    permissions = Set(title="Permissions",
                      description="ID of role's permissions",
                      value_type=TextLine(),
                      required=False)

    managers = Set(title="Managers",
                   description="List of principal IDs allowed to manage this role. "
                               "If it's a role, use 'role:role_id' syntax...",
                   value_type=TextLine(),
                   required=False)


class IPrincipalInfo(Interface):
    """Principal info class

    This is the generic interface of objects defined in request 'principal' attribute.
    """

    # pylint: disable=invalid-name
    id = TextLine(title="Globally unique ID",
                  required=True)

    title = TextLine(title="Principal name",
                     required=True)

    attributes = Dict(title="Principal groups",
                      description="IDs of principals to which this principal directly belongs",
                      value_type=TextLine())



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.plugin.userfolder module

This module is used to define local users.
"""

import base64
import hashlib
import hmac
import logging
import random
import sys
from datetime import datetime
from os import urandom

from persistent import Persistent
from pyramid.renderers import render
from pyramid_mailer.message import Attachment, Message
from zope.component.interfaces import ISite
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.interface import Invalid
from zope.password.interfaces import IPasswordManager
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_i18n.interfaces import II18n
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.plugin import ILocalUser, IUsersFolderPlugin, SALT_SIZE
from pyams_security.interfaces.base import IPrincipalInfo
from pyams_security.interfaces.names import PRINCIPAL_ID_FORMATTER, UNCHANGED_PASSWORD, \
    USERS_FOLDERS_VOCABULARY_NAME
from pyams_security.interfaces.notification import INotificationSettings
from pyams_security.principal import PrincipalInfo
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.factory import factory_config
from pyams_utils.html import html_to_text
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.request import check_request
from pyams_utils.traversing import get_parent
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'

from pyams_security import _  # pylint: disable=ungrouped-imports


LOGGER = logging.getLogger('PyAMS (security)')


#
# Users folder
#

@factory_config(IUsersFolderPlugin)
class UsersFolder(Folder):
    """Local users folder"""

    prefix = FieldProperty(IUsersFolderPlugin['prefix'])
    title = FieldProperty(IUsersFolderPlugin['title'])
    enabled = FieldProperty(IUsersFolderPlugin['enabled'])

    def authenticate(self, credentials, request):  # pylint: disable=unused-argument
        """Try to authenticate given credentials"""
        if not self.enabled:
            return None
        attrs = credentials.attributes
        login = attrs.get('login')
        principal = self.get(login)
        if principal is not None:
            password = attrs.get('password')
            if principal.check_password(password):
                return PRINCIPAL_ID_FORMATTER.format(prefix=self.prefix,
                                                     login=principal.login)
        return None

    def check_login(self, login):
        """Check if provided login is not already used"""
        if not login:
            return False
        return login not in self

    def get_principal(self, principal_id, info=True):
        """Get principal info matching given principal ID"""
        if not self.enabled:
            return None
        if not principal_id.startswith(self.prefix + ':'):
            return None
        prefix, login = principal_id.split(':', 1)  # pylint: disable=unused-variable
        user = self.get(login)
        if user is not None:
            if info:
                return PrincipalInfo(id=PRINCIPAL_ID_FORMATTER.format(prefix=self.prefix,
                                                                      login=user.login),
                                     title=user.title)
        return user

    def get_all_principals(self, principal_id):
        """Get all principals for given principal ID"""
        if not self.enabled:
            return set()
        if self.get_principal(principal_id) is not None:
            return {principal_id}
        return set()

    def find_principals(self, query, exact_match=False):
        """Get iterator of principals matching given query"""
        if not self.enabled:
            return
        if not query:
            return
        query = query.lower()
        for user in self.values():
            if exact_match:
                attrs = (user.login,)
            else:
                attrs = (user.login, user.firstname, user.lastname, user.email)
            for attr in attrs:
                if not attr:
                    continue
                if (exact_match and query == attr.lower()) or \
                        (not exact_match and query in attr.lower()):
                    yield PrincipalInfo(id=PRINCIPAL_ID_FORMATTER.format(prefix=self.prefix,
                                                                         login=user.login),
                                        title='{title} <{email}>'.format(title=user.title,
                                                                         email=user.email))
                    break

    def get_search_results(self, data):
        """Search iterator of principals matching given data"""
        query = data.get('query')
        if not query:
            return
        query = query.lower()
        for user in self.values():
            if (query == user.login or
                    query in user.firstname.lower() or
                    query in user.lastname.lower()):
                yield user


@vocabulary_config(name=USERS_FOLDERS_VOCABULARY_NAME)
class UsersFolderVocabulary(SimpleVocabulary):
    """'PyAMS users folders' vocabulary"""

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        terms = []
        manager = query_utility(ISecurityManager)
        if manager is not None:
            for name, plugin in manager.items():
                if IUsersFolderPlugin.providedBy(plugin):
                    terms.append(SimpleTerm(name, title=plugin.title))
        super().__init__(terms)


#
# Local users
#

def notify_user_activation(user, request=None):
    """Send mail for user activation"""
    security = query_utility(ISecurityManager)
    settings = INotificationSettings(security)
    if not settings.enable_notifications:  # pylint: disable=assignment-from-no-return
        LOGGER.info("Security notifications disabled, no message sent...")
        return
    mailer = settings.get_mailer()  # pylint: disable=assignment-from-no-return
    if mailer is None:
        LOGGER.warning("Can't find mailer utility, no notification message sent!")
        return
    if request is None:
        request = check_request()
    translate = request.localizer.translate
    i18n_settings = II18n(settings)
    message_text, template_name = None, None
    if user.self_registered:
        # pylint: disable=assignment-from-no-return
        message_text = i18n_settings.query_attribute('registration_template', request=request)
        if not message_text:
            template_name = 'templates/register-message.pt'
    elif user.wait_confirmation:
        # pylint: disable=assignment-from-no-return
        message_text = i18n_settings.query_attribute('confirmation_template', request=request)
        if not message_text:
            template_name = 'templates/register-info.pt'
    site = get_parent(request.context, ISite)
    if message_text is not None:
        message_text = message_text.format(**user.to_dict())
    elif template_name is not None:
        message_text = render(template_name, request=request, value={
            'user': user,
            'site': site,
            'settings': settings
        })
    html_body = render('templates/register-body.pt', request=request, value={
        'user': user,
        'site': site,
        'settings': settings,
        'message': message_text
    })
    message = Message(
        subject=translate(_("{prefix}Please confirm registration")).format(
            prefix="{prefix} ".format(prefix=settings.subject_prefix)
            if settings.subject_prefix else ''),
        sender='{name} <{email}>'.format(name=settings.sender_name,
                                         email=settings.sender_email),
        recipients=("{firstname} {lastname} <{email}>".format(firstname=user.firstname,
                                                              lastname=user.lastname,
                                                              email=user.email),),
        html=Attachment(data=html_body,
                        content_type='text/html; charset=utf-8',
                        disposition='inline',
                        transfer_encoding='quoted-printable'),
        body=Attachment(data=html_to_text(html_body),
                        content_type='text/plain; charset=utf-8',
                        disposition='inline',
                        transfer_encoding='quoted-printable'))
    mailer.send(message)


@factory_config(ILocalUser)
class LocalUser(Persistent, Contained):
    # pylint: disable=too-many-instance-attributes
    """Local user persistent class"""

    login = FieldProperty(ILocalUser['login'])
    email = FieldProperty(ILocalUser['email'])
    firstname = FieldProperty(ILocalUser['firstname'])
    lastname = FieldProperty(ILocalUser['lastname'])
    company_name = FieldProperty(ILocalUser['company_name'])
    password_manager = FieldProperty(ILocalUser['password_manager'])
    _password = FieldProperty(ILocalUser['password'])
    _password_salt = None
    wait_confirmation = FieldProperty(ILocalUser['wait_confirmation'])
    self_registered = FieldProperty(ILocalUser['self_registered'])
    activation_secret = FieldProperty(ILocalUser['activation_secret'])
    activation_hash = FieldProperty(ILocalUser['activation_hash'])
    activation_date = FieldProperty(ILocalUser['activation_date'])
    activated = FieldProperty(ILocalUser['activated'])

    @property
    def title(self):
        """Concatenate first and last names"""
        return '{firstname} {lastname}'.format(firstname=self.firstname,
                                               lastname=self.lastname)

    @property
    def password(self):
        """Get current encoded password"""
        return self._password

    @password.setter
    def password(self, value):
        """Encode and set user password"""
        if value:
            if value == UNCHANGED_PASSWORD:
                return
            self._password_salt = urandom(SALT_SIZE.get(self.password_manager, 4))
            manager = get_utility(IPasswordManager, name=self.password_manager)
            if self.password_manager == 'Plain Text':
                self._password = manager.encodePassword(value)
            else:
                self._password = manager.encodePassword(value, salt=self._password_salt)
        else:
            self._password = None

    def check_password(self, password):
        """Check given password with encoded one"""
        if not (self.activated and self.password):
            return False
        manager = query_utility(IPasswordManager, name=self.password_manager)
        if manager is None:
            return False
        return manager.checkPassword(self.password, password)

    def generate_secret(self, notify=True, request=None):
        """Generate activation secret and activation hash"""
        seed = self.activation_secret = '-'.join((hex(random.randint(0, sys.maxsize))[2:]
                                                  for i in range(5)))
        secret = hmac.new(self.password or b'', self.login.encode(), digestmod=hashlib.sha256)
        secret.update(seed.encode())
        self.activation_hash = base64.b32encode(secret.digest()).decode()
        if notify:
            notify_user_activation(self, request)

    def refresh_secret(self, notify=True, request=None):
        """Refresh activation secret and activation hash"""
        self.password = None
        self.generate_secret(notify, request)
        self.wait_confirmation = True
        self.activation_date = None
        self.activated = False

    def check_activation(self, hash, login, password):  # pylint: disable=redefined-builtin
        """Check is given hash is matching stored one, and activate user"""
        if self.self_registered:
            # If principal was registered by it's own, we check activation hash
            # with given login and password
            manager = get_utility(IPasswordManager, name=self.password_manager)
            password = manager.encodePassword(password, salt=self._password_salt)
            secret = hmac.new(password, login.encode(), digestmod=hashlib.sha256)
            secret.update(self.activation_secret.encode())
            activation_hash = base64.b32encode(secret.digest()).decode()
            if hash != activation_hash:
                raise Invalid(_("Can't activate profile with given params!"))
        else:
            # If principal was registered by a site manager, just check that
            # hash is matching stored one and update user password...
            if hash != self.activation_hash:
                raise Invalid(_("Can't activate profile with given params!"))
            self.password = password
        self.wait_confirmation = False
        self.activation_date = datetime.utcnow()
        self.activated = True

    def to_dict(self):
        """Get main user properties as mapping"""
        return {
            'login': self.login,
            'email': self.email,
            'firstname': self.firstname,
            'lastname': self.lastname,
            'title': self.title,
            'company_name': self.company_name
        }


@adapter_config(required=ILocalUser, provides=IPrincipalInfo)
def user_principal_info_adapter(user):
    """User principal info adapter"""
    return PrincipalInfo(id=PRINCIPAL_ID_FORMATTER.format(prefix=user.__parent__.prefix,
                                                          login=user.login),
                         title=user.title)


try:
    from pyams_mail.interfaces import IPrincipalMailInfo
except ImportError:
    pass
else:
    @adapter_config(required=ILocalUser, provides=IPrincipalMailInfo)
    class UserPrincipalMailInfoAdapter(ContextAdapter):
        """User principal mail info adapter"""

        def get_addresses(self):
            """Get user email address"""
            yield ('{0} {1}'.format(self.context.firstname,
                                    self.context.lastname),
                   self.context.email)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.credential module

This module defines credentials class.
"""

from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces.plugin import ICredentials


__docformat__ = 'restructuredtext'


@implementer(ICredentials)
class Credentials:
    """Credentials class"""

    prefix = FieldProperty(ICredentials['prefix'])
    id = FieldProperty(ICredentials['id'])  # pylint: disable=invalid-name
    attributes = FieldProperty(ICredentials['attributes'])

    def __init__(self, prefix, id, **attributes):  # pylint: disable=invalid-name,redefined-builtin
        self.prefix = prefix
        self.id = id  # pylint: disable=invalid-name
        self.attributes.update(**attributes)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.generations main module

"""

from importlib import import_module

import sys
from zope.lifecycleevent import ObjectCreatedEvent
from zope.principalannotation.interfaces import IPrincipalAnnotationUtility
from zope.principalannotation.utility import PrincipalAnnotationUtility

from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.names import ADMIN_USER_LOGIN, INTERNAL_USER_LOGIN, \
    INTERNAL_USER_NAME
from pyams_security.interfaces.names import ADMIN_USER_NAME, SYSTEM_PREFIX
from pyams_security.plugin.admin import AdminAuthenticationPlugin
from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_utils.registry import get_current_registry, utility_config


__docformat__ = 'restructuredtext'


RENAMED_CLASSES = {
    'pyams_security.interfaces ISocialUsersFolderPlugin':
        'pyams_auth_oauth.interfaces IOAuthUsersFolderPlugin',
    'pyams_security.interfaces ISocialUser':
        'pyams_auth_oauth.interfaces IOAuthUser',
    'pyams_security.interfaces ISocialLoginProviderInfo':
        'pyams_auth_oauth.interfaces IOAuthLoginProviderInfo',
    'pyams_security.interfaces ISocialLoginConfiguration':
        'pyams_auth_oauth.interfaces IOAuthLoginConfiguration',
    'pyams_security.interfaces ISocialLoginProviderConnection':
        'pyams_auth_oauth.interfaces IOAuthLoginProviderConnection',
    'pyams_security.plugin.social SocialUser':
        'pyams_auth_oauth.plugin OAuthUser',
    'pyams_security.plugin.social SocialUsersFolder':
        'pyams_auth_oauth.plugin OAuthUsersFolder',
    'pyams_security.plugin.social SocialLoginProviderInfo':
        'pyams_auth_oauth.plugin OAuthLoginProviderInfo',
    'pyams_security.plugin.social SocialLoginConfiguration':
        'pyams_auth_oauth.plugin OAuthLoginConfiguration',
    'pyams_security.plugin.social SocialLoginProviderConnection':
        'pyams_auth_oauth.plugin OAuthLoginProviderConnection'
}


REQUIRED_UTILITIES = (
    (ISecurityManager, '', None, 'Security manager'),
    (IPrincipalAnnotationUtility, '', PrincipalAnnotationUtility, 'User profiles')
)


def get_admin_user():
    """Get system manager profile"""
    admin_auth = AdminAuthenticationPlugin()
    admin_auth.prefix = SYSTEM_PREFIX
    admin_auth.title = 'System manager authentication'
    admin_auth.login = ADMIN_USER_LOGIN
    admin_auth.password = 'admin'
    return admin_auth


def get_service_user():
    """Get internal services profile"""
    service_auth = AdminAuthenticationPlugin()
    service_auth.prefix = SYSTEM_PREFIX
    service_auth.title = 'internal service'
    service_auth.login = INTERNAL_USER_LOGIN
    service_auth.password = None
    return service_auth


@utility_config(name='PyAMS security', provides=ISiteGenerations)
class SecurityGenerationsChecker:
    """I18n generations checker"""

    order = 50
    generation = 2

    def evolve(self, site, current=None):  # pylint: disable=no-self-use,unused-argument
        """Check for required utilities"""
        check_required_utilities(site, REQUIRED_UTILITIES)
        manager = site.getSiteManager().queryUtility(ISecurityManager)
        if manager is not None:
            if ADMIN_USER_NAME not in manager:
                admin_auth = get_admin_user()
                get_current_registry().notify(ObjectCreatedEvent(admin_auth))
                manager[ADMIN_USER_NAME] = admin_auth
            if INTERNAL_USER_NAME not in manager:
                service_auth = get_service_user()
                get_current_registry().notify(ObjectCreatedEvent(service_auth))
                manager[INTERNAL_USER_NAME] = service_auth
        if not current:
            current = 1
        for generation in range(current, self.generation):
            module_name = 'pyams_security.generations.evolve{}'.format(generation)
            module = sys.modules.get(module_name)
            if module is None:
                module = import_module(module_name)
            module.evolve(site)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.permission module

This module provides classes related to permissions definition and registration.
"""

from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_security.interfaces import IViewContextPermissionChecker
from pyams_security.interfaces.base import IPermission
from pyams_security.interfaces.names import PERMISSIONS_VOCABULARY_NAME
from pyams_utils.adapter import query_adapter
from pyams_utils.request import check_request
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'


@implementer(IPermission)
class Permission:
    """Permission utility class"""

    id = FieldProperty(IPermission['id'])  # pylint: disable=invalid-name
    title = FieldProperty(IPermission['title'])
    description = FieldProperty(IPermission['description'])

    def __init__(self, values=None, **args):  # pylint: disable=unused-argument
        if not isinstance(values, dict):
            values = args
        self.id = values.get('id')  # pylint: disable=invalid-name
        self.title = values.get('title')
        self.description = values.get('description')


def register_permission(config, permission):
    """Register a new permission

    Permissions registry is not required.
    But only registered permissions can be applied via default
    ZMI features.

    :param config: Pyramid configuration
    :param permission: permission attributes mapping
    """
    if not IPermission.providedBy(permission):
        if isinstance(permission, dict):
            permission = Permission(id=permission.get('id'),
                                    title=permission.get('title'))
        else:
            permission = Permission(id=permission, title=permission)
    config.registry.registerUtility(permission, IPermission, name=permission.id)


@vocabulary_config(name=PERMISSIONS_VOCABULARY_NAME)
class PermissionsVocabulary(SimpleVocabulary):
    """Permissions vocabulary"""

    interface = IPermission

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        request = check_request()
        registry = request.registry
        translate = request.localizer.translate
        terms = [SimpleTerm(p.id, title=translate(p.title))
                 for n, p in registry.getUtilitiesFor(self.interface)]
        terms.sort(key=lambda x: x.title)
        super().__init__(terms)


def get_permission_checker(request, context=None, view=None, action=''):
    """Get permission checker for given action

    :param request: current request
    :param context: current context
    :param view: current view
    :param action: name of the action to be checked; this may be the name of a custom
        adapter registered for :ref:`IViewContextPermissionChecker`
    """
    return query_adapter(IViewContextPermissionChecker, request, context, view, action)


def get_edit_permission(request, context=None, view=None, action=''):
    """Get required edit permission

    :param request: current request
    :param context: current context
    :param view: current view
    :param action: name of the action to be checked; this may be the name of a custom
        adapter registered for :ref:`IViewContextPermissionChecker`. If no permission
        checker is available for given action, a registry lookup will be made for
        default one.
    """
    checker = get_permission_checker(request, context, view, action)
    if (checker is None) and action:
        checker = get_permission_checker(request, context, view)
    if checker is not None:
        return checker.edit_permission
    return None



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces.rest module

This module defines interfaces of components which are used to improve security of
REST services using CORS requests.
"""

from zope.interface import Interface
from zope.schema import Bool

from pyams_utils.schema import TextLineListField


__docformat__ = 'restructuredtext'

from pyams_security import _


CORS_CONFIGURATION_KEY = 'pyams_security.cors'
"""CORS configuration annotation key"""


class ICORSSecurityInfo(Interface):
    """Base CORS security interface"""

    restrict_origins = Bool(title=_("Check request origin"),
                            description=_("If enabled, REST API will check requests against "
                                          "currently allowed origins, and raise an HTTP "
                                          "forbidden exception if not allowed"),
                            required=True,
                            default=True)

    allowed_origins = TextLineListField(title=_("Allowed origins"),
                                        description=_("List of allowed origins URLs using CORS "
                                                      "requests"),
                                        required=False)

    def check_origin(self, request):
        """Check origin of given request

        Raise an HTTP exception if request origin is not allowed.
        """

    def set_headers(self, request, allowed_methods=None):
        """Set CORS headers of given request"""



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.plugin.admin module

This module defines system principals which are used for system management tasks and for
internal services.
"""

from os import urandom

from persistent import Persistent
from zope.container.contained import Contained
from zope.interface import implementer
from zope.password.interfaces import IPasswordManager
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces.names import UNCHANGED_PASSWORD
from pyams_security.interfaces.plugin import IAdminAuthenticationPlugin, IDirectoryPlugin
from pyams_security.principal import PrincipalInfo
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_utility


__docformat__ = 'restructuredtext'


@factory_config(IAdminAuthenticationPlugin)
@implementer(IDirectoryPlugin)
class AdminAuthenticationPlugin(Persistent, Contained):
    """Hard-coded administrator authenticator plug-in

    This plug-in should only be enabled in development mode!!!
    """

    prefix = FieldProperty(IAdminAuthenticationPlugin['prefix'])
    title = FieldProperty(IAdminAuthenticationPlugin['title'])
    enabled = FieldProperty(IAdminAuthenticationPlugin['enabled'])

    login = FieldProperty(IAdminAuthenticationPlugin['login'])
    _password = FieldProperty(IAdminAuthenticationPlugin['password'])
    _password_salt = None

    @property
    def password(self):
        """Get current password"""
        return self._password

    @password.setter
    def password(self, value):
        """Encode password before storing new value"""
        if value:
            if value == UNCHANGED_PASSWORD:
                return
            self._password_salt = urandom(4)
            manager = get_utility(IPasswordManager, name='SSHA')
            self._password = manager.encodePassword(value, salt=self._password_salt)
        else:
            self._password = None

    def authenticate(self, credentials, request):  # pylint: disable=unused-argument
        """Try to authenticate principal using given credentials"""
        if not (self.enabled and self.password):
            return None
        attrs = credentials.attributes
        login = attrs.get('login')
        password = attrs.get('password')
        manager = get_utility(IPasswordManager, name='SSHA')
        if login == self.login and manager.checkPassword(self.password, password):
            return "{0}:{1}".format(self.prefix, login)
        return None

    def get_principal(self, principal_id, info=True):
        """Get principal matching given principal ID"""
        if not self.enabled:
            return None
        if not principal_id.startswith(self.prefix + ':'):
            return None
        prefix, login = principal_id.split(':', 1)
        if (prefix == self.prefix) and (login == self.login):
            if info:
                return PrincipalInfo(id=principal_id,
                                     title=self.title)
            return self
        return None

    def get_all_principals(self, principal_id):
        """Get all principals matching given principal ID"""
        if not self.enabled:
            return set()
        if self.get_principal(principal_id) is not None:
            return {principal_id}
        return set()

    def find_principals(self, query, exact_match=False):
        """Search principals matching given query"""
        if not query:
            return
        query = query.lower()
        title = self.title.lower()
        if (query == self.login) or (not exact_match and query in title):
            yield PrincipalInfo(id='{0}:{1}'.format(self.prefix, self.login),
                                title=self.title)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.include module

This module is used for Pyramid integration.
"""

from pyramid.settings import asbool
from zope.interface import classImplements
from zope.password.interfaces import IPasswordManager
from zope.password.password import MD5PasswordManager, PlainTextPasswordManager, \
    SHA1PasswordManager, SSHAPasswordManager

from pyams_security.interfaces import IDefaultProtectionPolicy
from pyams_security.interfaces.base import MANAGE_PERMISSION, MANAGE_ROLES_PERMISSION, \
    MANAGE_SECURITY_PERMISSION, MANAGE_SYSTEM_PERMISSION, PUBLIC_PERMISSION, \
    ROLE_ID, VIEW_PERMISSION, VIEW_SYSTEM_PERMISSION
from pyams_security.interfaces.names import ADMIN_USER_ID, SYSTEM_ADMIN_ROLE, SYSTEM_VIEWER_ROLE
from pyams_security.permission import register_permission
from pyams_security.plugin import PluginSelector
from pyams_security.role import RoleSelector, register_role, upgrade_role
from pyams_security.security import ProtectedObjectMixin
from pyams_security.utility import get_principal
from pyams_site.site import BaseSiteRoot


__docformat__ = 'restructuredtext'

from pyams_security import _  # pylint: disable=ungrouped-imports


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_security:locales')

    config.registry.registerUtility(factory=PlainTextPasswordManager,
                                    provided=IPasswordManager, name='Plain Text')
    config.registry.registerUtility(factory=MD5PasswordManager,
                                    provided=IPasswordManager, name='MD5')
    config.registry.registerUtility(factory=SHA1PasswordManager,
                                    provided=IPasswordManager, name='SHA1')
    config.registry.registerUtility(factory=SSHAPasswordManager,
                                    provided=IPasswordManager, name='SSHA')

    # add configuration directives
    config.add_directive('register_permission', register_permission)
    config.add_directive('register_role', register_role)
    config.add_directive('upgrade_role', upgrade_role)

    # add request methods
    config.add_request_method(get_principal, 'principal', reify=True)

    # add subscribers predicate
    config.add_subscriber_predicate('role_selector', RoleSelector)
    config.add_subscriber_predicate('plugin_selector', PluginSelector)

    # register standard permissions
    config.register_permission({
        'id': PUBLIC_PERMISSION,
        'title': _("View public contents")
    })
    config.register_permission({
        'id': VIEW_PERMISSION,
        'title': _("View protected contents")
    })
    config.register_permission({
        'id': MANAGE_PERMISSION,
        'title': _("Manage contents properties")
    })
    config.register_permission({
        'id': VIEW_SYSTEM_PERMISSION,
        'title': _("View management screens")
    })
    config.register_permission({
        'id': MANAGE_SYSTEM_PERMISSION,
        'title': _("Manage system properties")
    })
    config.register_permission({
        'id': MANAGE_SECURITY_PERMISSION,
        'title': _("Manage security")
    })
    config.register_permission({
        'id': MANAGE_ROLES_PERMISSION,
        'title': _("Manage roles")
    })

    # register standard roles
    config.register_role({
        'id': SYSTEM_ADMIN_ROLE,
        'title': _("System manager (role)"),
        'permissions': {
            PUBLIC_PERMISSION, VIEW_PERMISSION, MANAGE_PERMISSION,
            MANAGE_SYSTEM_PERMISSION, VIEW_SYSTEM_PERMISSION, MANAGE_SECURITY_PERMISSION,
            MANAGE_ROLES_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE)
        }
    })
    config.register_role({
        'id': SYSTEM_VIEWER_ROLE,
        'title': _("System viewer (role)"),
        'permissions': {
            PUBLIC_PERMISSION, VIEW_PERMISSION, VIEW_SYSTEM_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE)
        }
    })

    # custom classes implementations
    if not asbool(config.registry.settings.get('pyams.security.disable-default-policy', False)):
        classImplements(BaseSiteRoot, IDefaultProtectionPolicy)
        BaseSiteRoot.__acl__ = ProtectedObjectMixin.__acl__

    config.scan()



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces.site module

This module defines site root roles interface.
"""

from pyams_security.interfaces import IContentRoles
from pyams_security.interfaces.names import SYSTEM_ADMIN_ROLE, SYSTEM_VIEWER_ROLE
from pyams_security.schema import PrincipalsSetField


__docformat__ = 'restructuredtext'

from pyams_security import _


class ISiteRootRoles(IContentRoles):
    """Site root roles"""

    managers = PrincipalsSetField(title=_("Site managers"),
                                  description=_("These principals are allowed to manage the "
                                                "whole application environment"),
                                  role_id=SYSTEM_ADMIN_ROLE,
                                  required=False)

    viewers = PrincipalsSetField(title=_("Site viewers"),
                                 description=_("These principals are allowed to view some "
                                               "application settings, without update"),
                                 role_id=SYSTEM_VIEWER_ROLE,
                                 required=False)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""
Generic test cases for pyams_security doctests
"""

import base64
import os
import sys

from pyramid.testing import DummyRequest, setUp
from pyramid.threadlocal import manager
from pyramid_mailer import IMailer
from pyramid_mailer.mailer import DummyMailer
from zope.interface import Interface

from pyams_i18n.attr import I18nAttributeAdapter
from pyams_i18n.interfaces import II18n
from pyams_security.interfaces import IRoleProtectedObject, \
    ISecurityManager
from pyams_security.interfaces.notification import INotificationSettings
from pyams_security.notification import NotificationSettings, security_notification_factory
from pyams_security.security import RoleProtectedObject
from pyams_utils.factory import register_factory
from pyams_utils.registry import set_local_registry


__docformat__ = 'restructuredtext'


def get_package_dir(value):
    """Get package directory"""
    package_dir = os.path.split(value)[0]
    if package_dir not in sys.path:
        sys.path.append(package_dir)
    return package_dir


def setup_tests_registry():
    """Initialize required adapters and utilities for testing"""
    config = setUp(hook_zca=True)
    registry = config.registry
    # register adapters
    registry.registerAdapter(I18nAttributeAdapter, (Interface, ), II18n)
    # register security factory
    register_factory(IRoleProtectedObject, RoleProtectedObject)
    # register notification components
    mailer = DummyMailer()
    register_factory(INotificationSettings, NotificationSettings)
    registry.registerUtility(mailer, provided=IMailer, name='mailer')
    registry.registerAdapter(security_notification_factory, (ISecurityManager, ),
                             provided=INotificationSettings)
    return config


def new_test_request(login, password, method='basic', context=None, registry=None):
    """Create test request with HTTP authorization header"""
    auth = base64.b64encode('{}:{}'.format(login, password).encode()).decode()
    request = DummyRequest(headers={'Authorization': '{} {}'.format(method, auth)},
                           context=context)
    if registry is not None:
        manager.clear()
        manager.push({
            'registry': registry,
            'request': request
        })
    return request



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

import logging

from ZODB.POSException import ConnectionStateError
from pyramid.authentication import AuthTktCookieHelper
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import Authenticated, Everyone
from zope.interface import implementer

from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.plugin import ICredentialsPlugin
from pyams_utils.registry import get_all_utilities_registered_for, query_utility
from pyams_utils.wsgi import wsgi_environ_cache


LOGGER = logging.getLogger('PyAMS (security)')


@implementer(IAuthenticationPolicy)
class PyAMSAuthenticationPolicy:
    """PyAMS authentication policy

    This authentication policy relies on a registered ISecurityManager utility.
    Use same authentication ticket as AuthTktAuthenticationPolicy.

    See `pyramid.authentication.AuthTktAuthenticationPolicy` to get description
    of other constructor arguments.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, secret,
                 cookie_name='auth_ticket',
                 secure=False,
                 include_ip=False,
                 timeout=None,
                 reissue_time=None,
                 max_age=None,
                 path="/",
                 http_only=False,
                 wild_domain=True,
                 hashalg='sha256',
                 parent_domain=False,
                 domain=None):
        self.cookie = AuthTktCookieHelper(secret,
                                          cookie_name=cookie_name,
                                          secure=secure,
                                          include_ip=include_ip,
                                          timeout=timeout,
                                          reissue_time=reissue_time,
                                          max_age=max_age,
                                          http_only=http_only,
                                          path=path,
                                          wild_domain=wild_domain,
                                          hashalg=hashalg,
                                          parent_domain=parent_domain,
                                          domain=domain)

    @property
    def credentials_plugins(self):
        """Get list of credentials plug-ins"""
        yield from get_all_utilities_registered_for(ICredentialsPlugin)

    @property
    def security_manager(self):
        """Get current security manager"""
        return query_utility(ISecurityManager)

    @wsgi_environ_cache('pyams_security.unauthenticated_userid', store_none=False)
    def unauthenticated_userid(self, request):
        """Get unauthenticated user ID from given request"""
        result = self.cookie.identify(request)
        if result:
            return result['userid']
        for plugin in self.credentials_plugins:
            credentials = plugin.extract_credentials(request)
            if credentials is not None:
                return credentials.id
        return None

    @wsgi_environ_cache('pyams_security.authenticated_userid', store_none=False)
    def authenticated_userid(self, request):
        """Get authenticated user ID from given request"""
        principal_id = self.unauthenticated_userid(request)
        if principal_id is None:
            return None
        try:
            manager = self.security_manager
            if manager is not None:
                return manager.authenticated_userid(request, principal_id)
        except ConnectionStateError:
            pass
        return None

    @wsgi_environ_cache('pyams_security.effective_principals')
    def effective_principals(self, request, context=None):
        """Get effective principals from given request"""
        try:
            LOGGER.debug(">>> getting principals for principal {0} ({1}) on {2!r}".format(
                request.principal.title, request.principal.id, context or request.context))
        except AttributeError:
            LOGGER.debug(">>> getting principals for request {0} on {1!r}".format(
                request, context or request.context))
        principals = {Everyone}
        principal_id = self.authenticated_userid(request)
        if principal_id:
            # get authenticated user principals
            principals.add(Authenticated)
            principals.add(principal_id)
            manager = self.security_manager
            if manager is not None:
                principals |= set(manager.effective_principals(principal_id, request, context))
        LOGGER.debug('<<< principals = {0}'.format(str(sorted(principals))))
        return principals

    def remember(self, request, principal, **kw):
        """Remember request authentication as cookie"""
        return self.cookie.remember(request, principal, **kw)

    def forget(self, request):
        """Reset authentication cookie"""
        return self.cookie.forget(request)



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

import re
from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import containers, contains
from zope.interface import Attribute, Interface, Invalid, implementer, invariant
from zope.location.interfaces import IContained
from zope.schema import Bool, Choice, Datetime, Dict, Set, Text, TextLine

from pyams_security.interfaces.names import PASSWORD_MANAGERS_VOCABULARY_NAME, USER_LOGIN_TITLE
from pyams_security.schema import PrincipalsSetField
from pyams_utils.schema import EncodedPasswordField


__docformat__ = 'restructuredtext'

from pyams_security import _


#
# Security plug-ins interfaces
#

class IPlugin(IContained, IAttributeAnnotatable):
    """Basic authentication plug-in interface"""

    containers('pyams_security.interfaces.IAuthentication')

    prefix = TextLine(title=_("Plug-in prefix"),
                      description=_("This prefix is mainly used by authentication plug-ins to "
                                    "mark principals"))

    title = TextLine(title=_("Plug-in title"),
                     required=False)

    enabled = Bool(title=_("Enabled plug-in?"),
                   description=_("You can choose to disable any plug-in..."),
                   required=True,
                   default=True)


class IPluginEvent(Interface):
    """Plug-in event interface"""

    plugin = Attribute("Event plug-in name")


#
# Credentials extraction plug-ins interfaces
#

class ICredentials(Interface):
    """Credentials interface"""

    prefix = TextLine(title="Credentials plug-in prefix",
                      description="Prefix of plug-in which extracted credentials")

    id = TextLine(title="Credentials ID")  # pylint: disable=invalid-name

    attributes = Dict(title="Credentials attributes",
                      description="Attributes dictionary defined by each credentials plug-in",
                      required=False,
                      default={})


class ICredentialsPluginInfo(Interface):
    """Credentials extraction plug-in base interface"""

    def extract_credentials(self, request):
        """Extract user credentials from given request

        Result of 'extract_credentials' call should be an ICredentials object for which
        id is the 'raw' principal ID (without prefix); only authentication plug-ins should
        add a prefix to principal IDs to distinguish principals
        """


class ICredentialsPlugin(ICredentialsPluginInfo, IPlugin):
    """Credentials extraction plug-in interface"""


#
# Authentication plug-ins interfaces
#

class IAuthenticationPluginInfo(Interface):
    """Principal authentication plug-in base interface"""

    def authenticate(self, credentials, request):
        """Authenticate given credentials and returns a principal ID or None"""


class IAuthenticationPlugin(IAuthenticationPluginInfo, IPlugin):
    """Principal authentication plug-in interface"""


ADMIN_AUTHENTICATION_PLUGIN_LABEL = _("System authentication plug-in")


class IAdminAuthenticationPlugin(IAuthenticationPlugin):
    """Admin authentication plug-in base interface"""

    login = TextLine(title=_("Admin. login"))

    password = EncodedPasswordField(title=_("Admin. password"),
                                    required=False)


class IAuthenticatedPrincipalEvent(IPluginEvent):
    """Authenticated principal event interface"""

    principal_id = Attribute("Authenticated principal ID")

    infos = Attribute("Event custom infos")


@implementer(IAuthenticatedPrincipalEvent)
class AuthenticatedPrincipalEvent:
    """Authenticated principal event"""

    def __init__(self, plugin, principal_id, **infos):
        self.plugin = plugin
        self.principal_id = principal_id
        self.infos = infos


#
# Directory plug-ins interfaces
#

class IDirectoryPluginInfo(Interface):
    """Principal directory plug-in interface"""

    def get_principal(self, principal_id, info=True):
        """Returns real principal matching given ID, or None

        If info is True, returns a PrincipalInfo record instead
        of original principal object
        """

    def get_all_principals(self, principal_id):
        """Returns all principals matching given principal ID"""

    def find_principals(self, query, exact_match=False):
        """Find principals matching given query

        Method may return an iterator
        """


class IDirectoryPlugin(IDirectoryPluginInfo, IPlugin):
    """Principal directory plug-in info"""


class IDirectorySearchPlugin(IDirectoryPlugin):
    """Principal directory plug-in supporting search"""

    def get_search_results(self, data):
        """Search principals matching given query data

        This method is used in back-office search views so may reply even
        when the plug-in is disabled.
        Method may return an iterator on his own content objects
        """


class IGroupsAwareDirectoryPlugin(Interface):
    """Marker interface for plug-ins handling groups"""


#
# User local registration
#

SALT_SIZE = {
    'SSHA512': 32,
    'PBKDF2': 32
}


USERS_FOLDER_PLUGIN_LABEL = _("Users folder plug-in")


class IUsersFolderPlugin(IAuthenticationPlugin, IDirectorySearchPlugin):
    """Local users folder interface"""

    contains('pyams_security.interfaces.ILocalUser')

    def check_login(self, login):
        """Check for existence of given login"""


MAJS = range(ord('A'), ord('Z') + 1)
MINS = range(ord('a'), ord('z') + 1)
NUMS = range(ord('0'), ord('9') + 1)


def check_password(password):
    """Check validity of a given password"""
    nbmaj = 0
    nbmin = 0
    nbn = 0
    nbo = 0
    for car in password:
        if ord(car) in MAJS:
            nbmaj += 1
        elif ord(car) in MINS:
            nbmin += 1
        elif ord(car) in NUMS:
            nbn += 1
        else:
            nbo += 1
    if [nbmin, nbmaj, nbn, nbo].count(0) > 1:
        raise Invalid(_("Your password must contain at least three of these kinds of characters: "
                        "lowercase letters, uppercase letters, numbers and special characters"))


EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")


class IUserRegistrationInfo(Interface):
    """User registration info"""

    login = TextLine(title=USER_LOGIN_TITLE,
                     description=_("If you don't provide a custom login, your login will be your "
                                   "email address..."),
                     required=False)

    @invariant
    def check_login(self):
        """Set login as mail when missing"""
        if not self.login:
            self.login = self.email

    email = TextLine(title=_("E-mail address"),
                     description=_("An email will be sent to this address to validate account "
                                   "activation; it will be used as your future user login"),
                     required=True)

    @invariant
    def check_email(self):
        """Check for valid email"""
        if not EMAIL_REGEX.match(self.email):
            raise Invalid(_("Your email address is not valid!"))

    firstname = TextLine(title=_("First name"),
                         required=True)

    lastname = TextLine(title=_("Last name"),
                        required=True)

    company_name = TextLine(title=_("Company name"),
                            required=False)

    password = EncodedPasswordField(title=_("Password"),
                                    description=_("Password must be at least 8 characters long, "
                                                  "and contain at least three kinds of "
                                                  "characters between lowercase letters, "
                                                  "uppercase letters, numbers and special "
                                                  "characters"),
                                    min_length=8,
                                    required=True)

    confirmed_password = EncodedPasswordField(title=_("Confirmed password"),
                                              required=True)

    @invariant
    def check_password(self):
        """Check for password confirmation"""
        if self.password != self.confirmed_password:
            raise Invalid(_("You didn't confirmed your password correctly!"))
        check_password(self.password)


class IUserRegistrationConfirmationInfo(Interface):
    """User registration confirmation info"""

    activation_hash = TextLine(title=_("Activation hash"),
                               required=True)

    login = TextLine(title=USER_LOGIN_TITLE,
                     required=True)

    password = EncodedPasswordField(title=_("Password"),
                                    min_length=8,
                                    required=True)

    confirmed_password = EncodedPasswordField(title=_("Confirmed password"),
                                              required=True)

    @invariant
    def check_password(self):
        """Check for password confirmation"""
        if self.password != self.confirmed_password:
            raise Invalid(_("You didn't confirmed your password correctly!"))
        check_password(self.password)


class ILocalUser(IAttributeAnnotatable):
    """Local user interface"""

    containers(IUsersFolderPlugin)

    login = TextLine(title=USER_LOGIN_TITLE,
                     required=True,
                     readonly=True)

    @invariant
    def check_login(self):
        """Set login as mail when missing"""
        if not self.login:
            self.login = self.email

    email = TextLine(title=_("User email address"),
                     required=True)

    @invariant
    def check_email(self):
        """Check for invalid email address"""
        if not EMAIL_REGEX.match(self.email):
            raise Invalid(_("Given email address is not valid!"))

    firstname = TextLine(title=_("First name"),
                         required=True)

    lastname = TextLine(title=_("Last name"),
                        required=True)

    title = Attribute("User full name")

    company_name = TextLine(title=_("Company name"),
                            required=False)

    password_manager = Choice(title=_("Password manager name"),
                              description=_("Utility used to encrypt user password"),
                              required=True,
                              vocabulary=PASSWORD_MANAGERS_VOCABULARY_NAME,
                              default='PBKDF2')

    password = EncodedPasswordField(title=_("Password"),
                                    min_length=8,
                                    required=False)

    confirmed_password = EncodedPasswordField(title=_("Confirmed password"),
                                              min_length=8,
                                              required=False)

    wait_confirmation = Bool(title=_("Wait confirmation?"),
                             description=_("If 'no', user will be activated immediately without "
                                           "waiting email confirmation"),
                             required=True,
                             default=True)

    @invariant
    def check_activated_user(self):
        """Check for missing password of activated user"""
        if not self.password and not self.wait_confirmation:
            raise Invalid(_("You can't activate an account without setting a password!"))

    self_registered = Bool(title=_("Self-registered profile?"),
                           required=True,
                           default=True,
                           readonly=True)

    activation_secret = TextLine(title=_("Activation secret key"),
                                 description=_("This private secret is used to create and check "
                                               "activation hash"))

    activation_hash = TextLine(title=_("Activation hash"),
                               description=_("This hash is provided into activation message URL. "
                                             "Activation hash is missing for local users which "
                                             "were registered without waiting their "
                                             "confirmation."))

    activated = Bool(title=_("Activated"),
                     required=True,
                     default=False)

    activation_date = Datetime(title=_("Activation date"),
                               required=False)

    def check_password(self, password):
        """Check user password against provided one"""

    def generate_secret(self, notify=True, request=None):
        """Generate secret key of this profile"""

    def refresh_secret(self, notify=True, request=None):
        """Refresh secret key of this profile"""

    def check_activation(self, hash, login, password):  # pylint: disable=redefined-builtin
        """Check activation for given settings"""

    def to_dict(self):
        """Get main user properties as mapping"""


#
# Principals groups
#

GROUPS_FOLDER_PLUGIN_LABEL = _("Groups folder plug-in")


class IGroupsFolderPlugin(IDirectorySearchPlugin, IGroupsAwareDirectoryPlugin):
    """Principals groups folder plug-in"""

    contains('pyams_security.interfaces.ILocalGroup')

    def check_group_id(self, group_id):
        """Check for existence of given group ID"""


class ILocalGroup(Interface):
    """Local principals group interface"""

    containers(IGroupsFolderPlugin)

    group_id = TextLine(title=_("Group ID"),
                        description=_("This ID should be unique between all groups"),
                        required=True,
                        readonly=True)

    title = TextLine(title=_("Title"),
                     description=_("Public label of this group"),
                     required=True)

    description = Text(title=_("Description"),
                       required=False)

    principals = PrincipalsSetField(title=_("Group principals"),
                                    description=_("IDs of principals contained in this group"),
                                    required=False,
                                    default=set())


class IPrincipalsGroupEvent(Interface):
    """Principals group event interface"""

    group = Attribute("Event source group")

    principals = Set(title="List of principals IDs",
                     value_type=TextLine())


class PrincipalsGroupEvent:
    """Principals group event"""

    def __init__(self, group, principals):
        self.group = group
        self.principals = principals


class IPrincipalsAddedToGroupEvent(IPrincipalsGroupEvent):
    """Interface of event fired when principals were added to group"""


@implementer(IPrincipalsAddedToGroupEvent)
class PrincipalsAddedToGroupEvent(PrincipalsGroupEvent):
    """Event fired when principals were added to group"""


class IPrincipalsRemovedFromGroupEvent(IPrincipalsGroupEvent):
    """Interface of event fired when principals were removed from group"""


@implementer(IPrincipalsRemovedFromGroupEvent)
class PrincipalsRemovedFromGroupEvent(PrincipalsGroupEvent):
    """Event fired when principals were removed from group"""



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.security module

This module provides role-based security policy and ACL management.

A few utility classes are also available to help using permissions on objects and
views objects.
"""

import logging

from persistent import Persistent
from persistent.dict import PersistentDict
from pyramid.decorator import reify
from pyramid.location import lineage
from pyramid.security import ALL_PERMISSIONS, Allow, Authenticated, DENY_ALL, Deny, Everyone
from zope.annotation import IAttributeAnnotatable
from zope.container.contained import Contained
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces import GrantedRoleEvent, IContentRoles, \
    IDefaultProtectionPolicy, IProtectedObject, IRoleProtectedObject, ISecurityContext, \
    RevokedRoleEvent
from pyams_security.interfaces.base import IPrincipalInfo, IRole, PUBLIC_PERMISSION, ROLE_ID
from pyams_security.interfaces.names import ADMIN_USER_ID
from pyams_security.permission import get_edit_permission
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_pyramid_registry, query_utility
from pyams_utils.request import check_request, request_property


__docformat__ = 'restructuredtext'

LOGGER = logging.getLogger('PyAMS (security)')


@factory_config(IRoleProtectedObject)
class RoleProtectedObject(Persistent, Contained):
    """Base class for object protected by roles"""

    inherit_parent_security = FieldProperty(IRoleProtectedObject['inherit_parent_security'])
    everyone_denied = FieldProperty(IRoleProtectedObject['everyone_denied'])
    everyone_granted = FieldProperty(IRoleProtectedObject['everyone_granted'])
    authenticated_denied = FieldProperty(IRoleProtectedObject['authenticated_denied'])
    authenticated_granted = FieldProperty(IRoleProtectedObject['authenticated_granted'])
    inherit_parent_roles = FieldProperty(IRoleProtectedObject['inherit_parent_roles'])

    def __init__(self):
        self._principals_by_role = PersistentDict()
        self._roles_by_principal = PersistentDict()

    def get_everyone_denied(self):
        """Get permissions denied to everyone"""
        permissions = self.everyone_denied or set()
        if self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = permissions | (protection.everyone_denied or set())
        return permissions

    def get_everyone_granted(self):
        """Get permissions granted to everyone"""
        permissions = self.everyone_granted or set()
        if self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = permissions | (protection.everyone_granted or set())
        return permissions

    def get_authenticated_denied(self):
        """Get permissions denied to authenticated users"""
        permissions = self.authenticated_denied or set()
        if self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = permissions | (protection.authenticated_denied or set())
        return permissions

    def get_authenticated_granted(self):
        """Get permissions granted to authenticated users"""
        permissions = self.authenticated_granted or set()
        if self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = permissions | (protection.authenticated_granted or set())
        return permissions

    def grant_role(self, role_id, principal_ids):
        """Grant role to selected principals"""
        registry = get_pyramid_registry()
        if IRole.providedBy(role_id):
            role_id = role_id.id
        if isinstance(principal_ids, str):
            principal_ids = {principal_ids}
        role_principals = self._principals_by_role.get(role_id) or set()
        for principal_id in principal_ids:
            if IPrincipalInfo.providedBy(principal_id):
                principal_id = principal_id.id
            if principal_id not in role_principals:
                principal_roles = self._roles_by_principal.get(principal_id) or set()
                role_principals.add(principal_id)
                principal_roles.add(role_id)
                self._roles_by_principal[principal_id] = principal_roles
                self._principals_by_role[role_id] = role_principals
                registry.notify(GrantedRoleEvent(self, role_id, principal_id))

    def revoke_role(self, role_id, principal_ids):
        """Revoke role to selected principals"""
        registry = get_pyramid_registry()
        if IRole.providedBy(role_id):
            role_id = role_id.id
        if isinstance(principal_ids, str):
            principal_ids = {principal_ids}
        role_principals = self._principals_by_role.get(role_id) or set()
        for principal_id in principal_ids.copy():
            if IPrincipalInfo.providedBy(principal_id):
                principal_id = principal_id.id
            if principal_id in role_principals:
                principal_roles = self._roles_by_principal.get(principal_id) or set()
                if principal_id in role_principals:
                    role_principals.remove(principal_id)
                if role_id in principal_roles:
                    principal_roles.remove(role_id)
                if principal_roles:
                    self._roles_by_principal[principal_id] = principal_roles
                elif principal_id in self._roles_by_principal:
                    del self._roles_by_principal[principal_id]
                if role_principals:
                    self._principals_by_role[role_id] = role_principals
                elif role_id in self._principals_by_role:
                    del self._principals_by_role[role_id]
                registry.notify(RevokedRoleEvent(self, role_id, principal_id))

    def get_principals(self, role_id):
        """Get principals which have selected role granted"""
        if IRole.providedBy(role_id):
            role_id = role_id.id
        return self._principals_by_role.get(role_id) or set()

    def get_roles(self, principal_id):
        """Get roles for given principal"""
        if IPrincipalInfo.providedBy(principal_id):
            principal_id = principal_id.id
        return self._roles_by_principal.get(principal_id) or set()

    def get_permissions(self, principal_id):
        """Get permissions for given principal"""
        registry = get_pyramid_registry()
        result = set()
        for role_id in self.get_roles(principal_id):
            role = registry.queryUtility(IRole, role_id)
            result |= role.permissions or set()
        return result

    def get_granted_roles(self):
        """Get granted roles on current context or parents"""
        roles = set(self._principals_by_role.keys())
        if self.inherit_parent_roles:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    roles = roles | protection.get_granted_roles()
        return roles

    @request_property(key=None)
    def __acl__(self):
        """Get ACL for current context

        The result is stored into current request annotations, so it's not supposed to change
        during request lifetime.
        """
        # always grant all permissions to system manager
        # and 'public' permission to everyone
        result = [(Allow, ADMIN_USER_ID, ALL_PERMISSIONS),
                  (Allow, Everyone, {PUBLIC_PERMISSION})]
        # grant access to all roles permissions
        for role_id in self.get_granted_roles():
            role = query_utility(IRole, role_id)
            if role is not None:
                result.append((Allow, ROLE_ID.format(role_id), role.permissions))
        # add denied permissions to everyone and authenticated
        permissions = self.get_everyone_denied()
        if permissions:
            result.append((Deny, Everyone, permissions))
        permissions = self.get_authenticated_denied()
        if permissions:
            result.append((Deny, Authenticated, permissions))
        # add allowed permissions to everyone and authenticated
        permissions = self.get_authenticated_granted()
        if permissions:
            result.append((Allow, Authenticated, permissions))
        permissions = self.get_everyone_granted()
        if permissions:
            result.append((Allow, Everyone, permissions))
        # deny all parent permissions if inheritance is disabled
        if not self.inherit_parent_security:
            result.append(DENY_ALL)
        LOGGER.debug('ACL({0!r}) = {1}'.format(self.__parent__, str(result)))
        return result


POLICY_ANNOTATIONS_KEY = 'pyams_security.policy'


@adapter_config(required=IDefaultProtectionPolicy,
                provides=IRoleProtectedObject)
def protected_object_factory(context):
    """Default protected object factory"""
    return get_annotation_adapter(context, POLICY_ANNOTATIONS_KEY, IRoleProtectedObject)


@implementer(IContentRoles)
class ProtectedObjectRoles:
    """Protected object roles base class"""

    def __init__(self, context):
        self.__parent__ = context


@implementer(IAttributeAnnotatable)
class ProtectedObjectMixin:
    """Base protected object class mixin

    This mixin class is only used to automatically extract ACLs from an :py:class:`IProtectedObject
    <pyams_security.interfaces.IProtectedObject>` interface adapter.
    """

    def __acl__(self):
        protected = IProtectedObject(self, None)
        if protected is not None:
            acl = protected.__acl__()  # pylint: disable=assignment-from-no-return
            if callable(acl):
                acl = acl(protected)
            return acl
        return []


class ProtectedViewObjectMixin:
    """Base protected view object mixin

    This mixin class can be used to create view objects which will use an adapter to get
    the permission required to render the component (which can be a viewlet, a table column
    or anything else using a *permission* property).

    Context is extracted using an adapter of current object to :py:class:`ISecurityContext
    <pyams_security.interfaces.ISecurityContext>`, if available.

    The class then relies on an adapter to :py:class:`IViewContextPermissionChecker
    <pyams_security.interfaces.IViewContextPermissionChecker>` to get this permission. If
    defined, the *action_type* property can be used to get a custom adapter registered with
    this name; if not adapter is registered with this name, the default one is used.
    """

    action_type = ''

    @reify
    def permission(self):
        """Object permission getter"""
        request = getattr(self, 'request', None) or check_request()
        context = ISecurityContext(self, None)
        if context is None:
            context = getattr(self, 'context', request.context)
        view = getattr(self, 'view', None) or getattr(self, 'table', None)
        return get_edit_permission(request, context, view, self.action_type)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.notification module

This module handles notificaiton settings.
"""

from persistent import Persistent
from pyramid_mailer import IMailer
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.notification import INotificationSettings
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import query_utility


__docformat__ = 'restructuredtext'


@factory_config(INotificationSettings)
class NotificationSettings(Persistent):
    """Notification settings"""

    enable_notifications = FieldProperty(INotificationSettings['enable_notifications'])
    mailer = FieldProperty(INotificationSettings['mailer'])
    service_name = FieldProperty(INotificationSettings['service_name'])
    service_owner = FieldProperty(INotificationSettings['service_owner'])
    sender_name = FieldProperty(INotificationSettings['sender_name'])
    sender_email = FieldProperty(INotificationSettings['sender_email'])
    subject_prefix = FieldProperty(INotificationSettings['subject_prefix'])
    confirmation_template = FieldProperty(INotificationSettings['confirmation_template'])
    registration_template = FieldProperty(INotificationSettings['registration_template'])
    signature = FieldProperty(INotificationSettings['signature'])

    def get_mailer(self):
        """Get mailer utility matching current selection"""
        if self.mailer is not None:
            return query_utility(IMailer, name=self.mailer)
        return None


NOTIFICATIONS_KEY = 'pyams_security.notifications'


@adapter_config(required=ISecurityManager, provides=INotificationSettings)
def security_notification_factory(context):
    """Security manager notifications factory adapter"""
    return get_annotation_adapter(context, NOTIFICATIONS_KEY, INotificationSettings, locate=False)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.password module

This module defines custom passwords managers.
"""

from base64 import standard_b64decode, standard_b64encode
from hashlib import pbkdf2_hmac, sha512
from hmac import compare_digest
from os import urandom

from zope.password.interfaces import IPasswordManager
from zope.password.password import _PrefixedPasswordManager, _encoder

from pyams_security.interfaces.plugin import SALT_SIZE
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


@utility_config(name='SSHA512',
                provides=IPasswordManager)
class SSHA512PasswordManager(_PrefixedPasswordManager):
    """SSHA512 password manager"""

    _prefix = b'{SSHA512}'

    def encodePassword(self, password, salt=None):  # pylint: disable=arguments-differ
        if salt is None:
            salt = urandom(SALT_SIZE.get('SSHA512', 32))
        elif not isinstance(salt, bytes):
            salt = _encoder(salt)
        hashcode = sha512(_encoder(password))
        hashcode.update(salt)
        return self._prefix + standard_b64encode(hashcode.digest() + salt)

    def checkPassword(self, encoded_password, password):
        # standard_b64decode() cannot handle unicode input string.
        encoded_password = _encoder(encoded_password)[9:]
        byte_string = standard_b64decode(encoded_password)
        salt = byte_string[64:]
        return compare_digest(encoded_password,
                              self.encodePassword(password, salt)[9:])


@utility_config(name='PBKDF2',
                provides=IPasswordManager)
class PBKDF2PasswordManager(_PrefixedPasswordManager):
    """PBKDF2-HMAC password manager"""

    _prefix = b'{PBKDF2}'

    def encodePassword(self, password, salt=None):  # pylint: disable=arguments-differ
        if salt is None:
            salt = urandom(SALT_SIZE.get('PBKDF2', 32))
        elif not isinstance(salt, bytes):
            salt = _encoder(salt)
        hashcode = pbkdf2_hmac('sha512', _encoder(password), salt, 100000)
        return self._prefix + standard_b64encode(hashcode + salt)

    def checkPassword(self, encoded_password, password):
        # standard_b64decode() cannot handle unicode input string.
        encoded_password = _encoder(encoded_password)[8:]
        byte_string = standard_b64decode(encoded_password)
        salt = byte_string[64:]
        return compare_digest(encoded_password,
                              self.encodePassword(password, salt)[8:])



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.plugin main module

This module enables a "plugin_selector" subscriber predicate, which can be used to
filter plugins events based on plug-in name or interface.
"""

__docformat__ = 'restructuredtext'


class PluginSelector:
    """Plug-in based event selector

    This selector can be used by subscriber to filter authentication
    events based on the name of the plug-in which fired the event.
    """

    def __init__(self, name, config):  # pylint: disable=unused-argument
        self.plugin = name

    def text(self):
        """Predicate text output"""
        return 'plugin_selector = %s' % str(self.plugin)

    phash = text

    def __call__(self, event):
        if isinstance(self.plugin, str):
            return event.plugin == self.plugin
        try:
            if self.plugin.providedBy(event.plugin):
                return True
        except (AttributeError, TypeError):
            if isinstance(event.plugin, self.plugin):
                return True
        return False



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces.notification module

This module defines notifications related interfaces.
"""

from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import Invalid, invariant
from zope.schema import Bool, Choice, Text, TextLine

from pyams_i18n.schema import I18nHTMLField
from pyams_mail.interfaces import MAILERS_VOCABULARY_NAME


__docformat__ = 'restructuredtext'

from pyams_security import _


class INotificationSettings(IAttributeAnnotatable):
    """Mailer notification settings interface"""

    enable_notifications = Bool(title=_("Enable notifications?"),
                                description=_("If 'no', mail notifications will be disabled"),
                                required=True,
                                default=False)

    mailer = Choice(title=_("Mailer utility"),
                    description=_("Mail delivery utility used to send notifications"),
                    required=False,
                    vocabulary=MAILERS_VOCABULARY_NAME)

    @invariant
    def check_mailer(self):
        """Check mailer to enable notifications"""
        if self.enable_notifications and not self.mailer:
            raise Invalid(_("Notifications can't be enabled without mailer utility"))

    def get_mailer(self):
        """Return mailer utility"""

    service_name = TextLine(title=_("Service name"),
                            description=_("Name of service as defined in registration mail "
                                          "subject"),
                            required=True)

    service_owner = TextLine(title=_("Service owner"),
                             description=_("Name of the entity providing this service, which "
                                           "will be visible in notifications messages"),
                             required=True)

    sender_name = TextLine(title=_("Sender name"),
                           description=_("Visible name of registration mail sender"),
                           required=True)

    sender_email = TextLine(title=_("Sender email"),
                            description=_("Email address of registration mail sender"),
                            required=True)

    subject_prefix = TextLine(title=_("Subject prefix"),
                              description=_("This prefix will be inserted into subject prefix of "
                                            "each notification message"),
                              required=False)

    confirmation_template = I18nHTMLField(
        title=_("Confirmation template"),
        description=_("This template will be used instead of default template to send "
                      "notification when a user is registered by a system administrator; you can "
                      "use some user properties into the message body, like: {login}, {email}, "
                      "{firstname}, {lastname}, {title} or {company_name}; message activation "
                      "link and footer are added automatically"),
        required=False)

    registration_template = I18nHTMLField(
        title=_("Registration template"),
        description=_("This template will be used instead of default template to send "
                      "notificaiton when a user is auto-registered; you can use some user "
                      "properties into the message body, like: {login}, {email}, {firstname}, "
                      "{lastname}, {title} or {company_name}; message activation link and footer "
                      "are added automatically"),
        required=False)

    signature = Text(title=_("Email signature"),
                     description=_("Text displayed in email footer"),
                     required=False)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.skin main module

This module provides principals search function.
"""

from pyramid.view import view_config

from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.base import VIEW_SYSTEM_PERMISSION
from pyams_utils.registry import query_utility


__docformat__ = 'restructuredtext'


@view_config(name='find-principals.json', permission=VIEW_SYSTEM_PERMISSION,
             renderer='json', xhr=True)
def find_principals(request):
    """Find all principals matching given query"""
    query = request.params.get('query')
    if not query:
        return []
    manager = query_utility(ISecurityManager)
    return [{
        'id': principal.id,
        'text': principal.title
    } for principal in manager.find_principals(query, exact_match=False)]



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.utility module

This module defines the SecurityManager utility and custom Pyramid authentication policy.
"""

import logging
from beaker.cache import cache_region
from pyramid.location import lineage
from zope.container.folder import Folder
from zope.schema.fieldproperty import FieldProperty

from pyams_security.interfaces import IProtectedObject, ISecurityManager
from pyams_security.interfaces.base import ROLE_ID
from pyams_security.interfaces.plugin import AuthenticatedPrincipalEvent, IAuthenticationPlugin, \
    ICredentialsPlugin, IDirectoryPlugin, IGroupsAwareDirectoryPlugin
from pyams_security.principal import MissingPrincipal, UnknownPrincipal
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_all_utilities_registered_for, get_utilities_for, \
    query_utility
from pyams_utils.request import check_request


__docformat__ = 'restructuredtext'

LOGGER = logging.getLogger('PyAMS (security)')


@factory_config(ISecurityManager)
class SecurityManager(Folder):
    """Security manager utility"""

    open_registration = FieldProperty(ISecurityManager['open_registration'])
    users_folder = FieldProperty(ISecurityManager['users_folder'])

    authentication_plugins_names = FieldProperty(ISecurityManager['authentication_plugins_names'])
    directory_plugins_names = FieldProperty(ISecurityManager['directory_plugins_names'])

    show_home_menu = FieldProperty(ISecurityManager['show_home_menu'])

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if IAuthenticationPlugin.providedBy(value):
            self.authentication_plugins_names += (key,)
        if IDirectoryPlugin.providedBy(value):
            self.directory_plugins_names += (key,)

    def __delitem__(self, key):
        super().__delitem__(key)
        if key in self.authentication_plugins_names:
            self.authentication_plugins_names = tuple(
                filter(lambda x: x != key, self.authentication_plugins_names))
        if key in self.directory_plugins_names:
            self.directory_plugins_names = tuple(
                filter(lambda x: x != key, self.directory_plugins_names))

    def get_plugin(self, name):
        """Lookup plugin for given name"""
        return query_utility(ICredentialsPlugin, name=name) or self.get(name)

    @property
    def credentials_plugins_names(self):
        """Get list of credentials plugins names"""
        yield from [name for name, plugin in get_utilities_for(ICredentialsPlugin) if name]

    @property
    def credentials_plugins(self):
        """Get list of credentials plug-ins"""
        yield from get_all_utilities_registered_for(ICredentialsPlugin)

    @property
    def authentication_plugins(self):
        """Get list of authentication plug-ins"""
        yield from get_all_utilities_registered_for(IAuthenticationPlugin)
        for name in self.authentication_plugins_names or ():
            plugin = self.get(name)
            if IAuthenticationPlugin.providedBy(plugin):
                yield plugin

    @property
    def directory_plugins(self):
        """Get list of directory plug-ins"""
        yield from get_all_utilities_registered_for(IDirectoryPlugin)
        for name in self.directory_plugins_names or ():
            plugin = self.get(name)
            if IDirectoryPlugin.providedBy(plugin):
                yield plugin

    @property
    def groups_directory_plugins(self):
        """Get list of groups plug-ins"""
        yield from get_all_utilities_registered_for(IGroupsAwareDirectoryPlugin)
        for name in self.directory_plugins_names or ():
            plugin = self.get(name)
            if IGroupsAwareDirectoryPlugin.providedBy(plugin):
                yield plugin

    # IAuthenticationInfo interface methods
    def extract_credentials(self, request, **kwargs):
        """Extract credentials from request"""
        for plugin in self.credentials_plugins:
            credentials = plugin.extract_credentials(request, **kwargs)
            if credentials:
                return credentials
        return None

    def authenticate(self, credentials, request):
        """Try to authenticate request with given credentials"""
        for plugin in self.authentication_plugins:
            try:
                principal_id = plugin.authenticate(credentials, request)
            except:  # pylint: disable=bare-except
                LOGGER.debug("Can't authenticate!", exc_info=True)
                continue
            else:
                if principal_id is not None:
                    request.registry.notify(
                        AuthenticatedPrincipalEvent(plugin.prefix, principal_id))
                    return principal_id
        return None

    def authenticated_userid(self, request, principal_id=None):
        """Extract authenticated user ID from request"""
        if principal_id is None:
            credentials = self.extract_credentials(request)
            if credentials is None:
                return None
            principal_id = self.authenticate(credentials, request)
        if principal_id is not None:
            principal = self.get_principal(principal_id)
            if principal is not None:
                return principal.id
        return None

    @cache_region('short', 'security_plugins_principals')
    def _get_plugins_principals(self, principal_id):
        """Extract all principals of given principal ID"""
        principals = set()
        # get direct principals
        for plugin in self.directory_plugins:
            principals |= set(plugin.get_all_principals(principal_id))
        # get indirect principals by searching groups members
        for principal in principals.copy():
            for plugin in self.groups_directory_plugins:
                principals |= set(plugin.get_all_principals(principal))
        return principals

    def effective_principals(self, principal_id, request=None, context=None):
        """Extratc effective principals of given principal ID"""
        # add principals extracted from security plug-ins
        principals = self._get_plugins_principals(principal_id)
        # add context roles granted to principal
        if context is None:
            if request is None:
                request = check_request()
            context = request.context
        if context is not None:
            for parent in lineage(context):
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    for principal in principals.copy():
                        principals |= set(map(ROLE_ID.format,
                                              protection.get_roles(principal)))
                    if not protection.inherit_parent_roles:
                        break
        return principals

    # IDirectoryPlugin interface methods
    @cache_region('long', 'security_plugins_principal')
    def get_principal(self, principal_id, info=True):
        """Principal lookup for given principal ID"""
        if not principal_id:
            return UnknownPrincipal
        for plugin in self.directory_plugins:
            try:
                principal = plugin.get_principal(principal_id, info)
            except:  # pylint: disable=bare-except
                LOGGER.debug("Can't get principal {0}!".format(principal_id), exc_info=True)
                continue
            else:
                if principal is not None:
                    return principal
        return MissingPrincipal(id=principal_id)

    def get_raw_principal(self, principal_id):
        """Principal lookup for given principal ID

        The method will always bypass the cache and return a *raw* principal, instead of a
        generic principal info, and None if the principal can't be found.
        """
        if not principal_id:
            return None
        for plugin in self.directory_plugins:
            try:
                principal = plugin.get_principal(principal_id, info=False)
            except:
                LOGGER.debug("Can't get principal {0}!".format(principal_id), exc_info=True)
                continue
            else:
                if principal is not None:
                    return principal
        return MissingPrincipal(id=principal_id)

    def get_all_principals(self, principal_id):
        """Get all principals of given principal ID"""
        principals = set()
        if principal_id:
            for plugin in self.directory_plugins:
                principals.update(plugin.get_all_principals(principal_id))
        return principals

    def find_principals(self, query, exact_match=False):
        """Find principals matching given query"""
        principals = set()
        for plugin in self.directory_plugins:
            try:
                principals |= set(plugin.find_principals(query, exact_match))
            except:  # pylint: disable=bare-except
                LOGGER.debug("Can't find principals!", exc_info=True)
                continue
        return sorted(principals, key=lambda x: x.title)


def get_principal(request, principal_id=None):
    """Get principal associated with given request"""
    manager = query_utility(ISecurityManager)
    if manager is not None:
        if principal_id is None:
            principal_id = request.authenticated_userid
        if principal_id:
            return manager.get_principal(principal_id)
        return UnknownPrincipal
    return None



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.profile module

This module defines classes and functions to handle public profiles.
"""

from persistent import Persistent
from pyramid.interfaces import IRequest
from pyramid.security import ALL_PERMISSIONS, Allow, Everyone
from pyramid.threadlocal import get_current_request
from zope.container.contained import Contained
from zope.interface import Interface
from zope.intid.interfaces import IIntIds
from zope.location import locate
from zope.traversing.interfaces import ITraversable

from pyams_file.property import FileProperty
from pyams_security.interfaces.base import IPrincipalInfo, PUBLIC_PERMISSION
from pyams_security.interfaces.names import ADMIN_USER_ID
from pyams_security.interfaces.profile import IPublicProfile, PUBLIC_PROFILE_KEY
from pyams_utils.adapter import ContextRequestAdapter, adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.interfaces.tales import ITALESExtension
from pyams_utils.registry import get_utility
from pyams_utils.request import check_request, query_request


__docformat__ = 'restructuredtext'


@factory_config(IPublicProfile)
class PublicProfile(Persistent, Contained):
    """Public profile persistent class"""

    avatar = FileProperty(IPublicProfile['avatar'])

    @staticmethod
    def __acl__():
        """Default profile ACL"""
        result = [(Allow, ADMIN_USER_ID, ALL_PERMISSIONS)]
        request = query_request()
        if request is not None:
            result.append((Allow, request.principal.id, ALL_PERMISSIONS))
        result.append((Allow, Everyone, PUBLIC_PERMISSION))
        return result


@adapter_config(required=Interface,
                provides=IPublicProfile)
def public_profile_factory(context):  # pylint: disable=unused-argument
    """Generic public profile factory

    Applied on any context, this adapter returns public profile associated
    with the current request principal.
    """
    request = check_request()
    return IPublicProfile(request.principal)


@adapter_config(required=IRequest,
                provides=IPublicProfile)
def request_profile_factory(request):
    """Request public profile factory"""
    return IPublicProfile(request.principal)


@adapter_config(required=IPrincipalInfo,
                provides=IPublicProfile)
def principal_profile_factory(principal):
    """Principal public profile factory adapter

    Public profile is stored using IPrincipalAnnotations utility (using the
    IPrincipalInfo to IAnnotations adapter defined into :py:mod:pyams_security.principal
    module).
    """

    def public_profile_callback(profile):
        """Public profile creation callback"""
        request = get_current_request()
        if request is not None:
            root = request.root
            intids = get_utility(IIntIds)
            locate(profile, root)  # avoid NotYet exception
            locate(profile, root, '++profile++{0}'.format(intids.register(profile)))

    return get_annotation_adapter(principal, PUBLIC_PROFILE_KEY, IPublicProfile,
                                  locate=False, callback=public_profile_callback)


@adapter_config(name='profile',
                required=(Interface, Interface),
                provides=ITraversable)
class ProfileTraverser(ContextRequestAdapter):
    """++profile++ namespace traverser"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Profile traverser"""
        if not name:
            return IPublicProfile(self.request.principal)
        intids = get_utility(IIntIds)
        profile = intids.queryObject(int(name))
        return IPublicProfile(profile, None)


@adapter_config(name='public_profile',
                required=(Interface, Interface),
                provides=ITALESExtension)
class PublicProfileExtension(ContextRequestAdapter):
    """public_profile TALES extension"""

    def render(self, request=None):
        """Render TALES extension"""
        if request is None:
            request = self.request
        return IPublicProfile(request)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.interfaces module

This module provides all security-related constants and interfaces.
"""

from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Attribute, Interface, Invalid, implementer, invariant
from zope.interface.interfaces import IObjectEvent, ObjectEvent
from zope.schema import Bool, Choice, TextLine, Tuple

from pyams_security.interfaces.names import PASSWORD_MANAGERS_VOCABULARY_NAME, \
    USERS_FOLDERS_VOCABULARY_NAME
from pyams_security.interfaces.plugin import IDirectoryPluginInfo, IPlugin, IPluginEvent
from pyams_security.schema import PermissionsSetField


__docformat__ = 'restructuredtext'

from pyams_security import _  # pylint: disable=ungrouped-imports


#
# Roles events interfaces
#

class IRoleEvent(IObjectEvent):
    """Base role event interface"""

    role_id = Attribute("Modified role ID")

    principal_id = Attribute("Modified principal ID")


@implementer(IRoleEvent)
class RoleEvent(ObjectEvent):
    """Base role event"""

    def __init__(self, object, role_id, principal_id):  # pylint: disable=redefined-builtin
        super().__init__(object)
        self.role_id = role_id
        self.principal_id = principal_id


class IGrantedRoleEvent(IRoleEvent):
    """Granted role event interface"""


@implementer(IGrantedRoleEvent)
class GrantedRoleEvent(RoleEvent):
    """Granted role event"""


class IRevokedRoleEvent(IRoleEvent):
    """Revoked role event interface"""


@implementer(IRevokedRoleEvent)
class RevokedRoleEvent(RoleEvent):
    """Revoked role interface"""


#
# Security manager
#

class ISecurityManager(IContainer, IDirectoryPluginInfo, IAttributeAnnotatable):
    """Authentication and principals management utility"""

    contains(IPlugin)

    open_registration = Bool(title=_("Enable free registration?"),
                             description=_("If 'Yes', any use will be able to create a new user "
                                           "account"),
                             required=False,
                             default=False)

    users_folder = Choice(title=_("Users folder"),
                          description=_("Name of users folder used to store registered principals"),
                          required=False,
                          vocabulary=USERS_FOLDERS_VOCABULARY_NAME)

    @invariant
    def check_users_folder(self):
        """Check for open registration"""
        if self.open_registration and not self.users_folder:
            raise Invalid(_("You can't activate open registration without selecting a users "
                            "folder"))

    credentials_plugins_names = Tuple(title=_("Credentials plug-ins"),
                                      description=_("These plug-ins are used to extract "
                                                    "credentials from an incoming request (the "
                                                    "Pyramid session plug-in is built-in!). "
                                                    "They can rely on authentication plug-ins "
                                                    "to ensure that provided credentials are "
                                                    "valid..."),
                                      value_type=TextLine(),
                                      readonly=True)

    authentication_plugins_names = Tuple(title=_("Authentication plug-ins"),
                                         description=_("The plug-ins can be used to check "
                                                       "extracted credentials against a local or "
                                                       "remote users database"),
                                         value_type=TextLine(),
                                         default=())

    directory_plugins_names = Tuple(title=_("Directory plug-ins"),
                                    description=_("The plug-in can be used to extract principals "
                                                  "information"),
                                    value_type=TextLine(),
                                    default=())

    def get_plugin(self, name):
        """Get plug-in matching given name"""

    credentials_plugins = Attribute("Iterator on registered credentials plug-ins")

    authentication_plugins = Attribute("Iterator on registered and local authentication plug-ins")

    directory_plugins = Attribute("Iterator on registered and local directory plug-ins")

    groups_directory_plugins = Attribute("Iterator on registered and local groups plug-ins")

    def effective_principals(self, principal_id, request=None, context=None):
        """Get effective principals of provided principal ID"""

    def get_principal(self, principal_id, info=True):
        """Principal lookup for provided principal ID"""

    def get_raw_principal(self, principal_id):
        """Principal lookup for provided principal ID

        Unlike *get_principal* method, this one will always return a
        *raw* principal, instead of a generic principal info, and *None*
        if the principal can't be found.
        """

    def get_all_principals(self, principal_id):
        """Get all principals of given principal ID"""

    show_home_menu = Bool(title=_("Access menu from home"),
                          description=_("If 'yes', a menu will be displayed to get access to "
                                        "security manager from site admin home page"),
                          required=True,
                          default=False)


LOGIN_REFERER_KEY = 'pyams_security.login.referer'
"""Key of request annotation used to store referer"""


#
# Protected objects interfaces
#

class IProtectedObject(IAttributeAnnotatable):
    """Protected object interface

    This is the only interface used by authorization policy.
    So you are free to implement custom protection mechanisms.
    """

    inherit_parent_security = Bool(title=_("Inherit parent security?"),
                                   description=_("Get access control entries (ACE) inherited "
                                                 "from parent levels"),
                                   required=True,
                                   default=True)

    everyone_denied = PermissionsSetField(title=_("Public denied permissions"),
                                          description=_("These permissions will be denied to all "
                                                        "users. Denied permissions take "
                                                        "precedence over granted ones."),
                                          required=False)

    everyone_granted = PermissionsSetField(title=_("Public granted permissions"),
                                           description=_("These permissions will be granted to "
                                                         "all users"),
                                           required=False)

    authenticated_denied = PermissionsSetField(title=_("Authenticated denied permissions"),
                                               description=_("These permissions will be denied "
                                                             "to authenticated users. Denied "
                                                             "permissions take precedence over "
                                                             "granted ones."),
                                               required=False)

    authenticated_granted = PermissionsSetField(title=_("Authenticated granted permissions"),
                                                description=_("These permissions will be granted "
                                                              "to authenticated users"),
                                                required=False)

    inherit_parent_roles = Bool(title=_("Inherit parent roles?"),
                                description=_("Get roles granted on parent levels"),
                                required=True,
                                default=True)

    def __acl__(self):
        """Object ACL"""

    def get_principals(self, role_id):
        """Get ID of principals who were granted given role

        May return an empty set when empty
        """

    def get_roles(self, principal_id):
        """Get ID of roles granted to given principal

        May return an empty set when empty
        """

    def get_roles_ids(self, principal_id):
        """Get ID of roles granted to given principal"""

    def get_permissions(self, principal_id):
        """Get ID of permissions granted to given principal"""

    def get_everyone_denied(self):
        """Get denied permissions for everyone, including inherited ones"""

    def get_everyone_granted(self):
        """Get granted permissions for everyone, including inherited ones"""

    def get_authenticated_denied(self):
        """Get denied permissions for authenticated, including inherited ones"""

    def get_authenticated_granted(self):
        """Get granted permissions for authenticated, including inherited ones"""

    def get_granted_roles(self):
        """Get all roles, including inherited ones"""


class IRoleProtectedObject(IProtectedObject):
    """Roles protected object interface"""

    def grant_role(self, role, principal_ids):
        """Grant given role to principals"""

    def revoke_role(self, role, principal_ids):
        """Revoke given role from principals"""


class IDefaultProtectionPolicy(Interface):
    """Marker interface for objects using default protection policy"""


class IContentRoles(Interface):
    """Base content roles interface"""


class IRolesPolicy(Interface):
    """Roles policy interface"""

    roles_interface = Attribute("Name of interface containing roles fields")

    weight = Attribute("Weight ordering attribute")


class ISecurityContext(Interface):
    """Security context getter interface

    This interface may be used as a simple context adapter to get object from
    which security permissions should be extracted.
    """


class IViewContextPermissionChecker(Interface):
    """Interface used to check access permissions on view context

    May be implemented as a context adapter, or as a (context, request, view)
    multi-adapter.
    """

    edit_permission = Attribute("Permission required to update form's content")
